from __future__ import annotations

from typing import TYPE_CHECKING, Dict

import numpy as np

# noinspection PyProtectedMember
from officialeye._api.context import Context

# noinspection PyProtectedMember
from officialeye._api.future import Future

# noinspection PyProtectedMember
from officialeye._api.image import Image

# noinspection PyProtectedMember
from officialeye._api.template.match import IMatch

# noinspection PyProtectedMember
from officialeye._api.template.supervision_result import ISupervisionResult
from officialeye._internal.api.interpret import template_interpret
from officialeye._internal.api_implementation import IApiInterfaceImplementation

# noinspection PyProtectedMember
from officialeye._internal.template.external_interpretation_result import ExternalInterpretationResult
from officialeye._internal.template.external_matching_result import ExternalMatchingResult
from officialeye._internal.template.external_template import ExternalTemplate

if TYPE_CHECKING:
    # noinspection PyProtectedMember
    from officialeye._api.image import IImage
    from officialeye._internal.template.internal_supervision_result import InternalSupervisionResult


class ExternalSupervisionResult(ISupervisionResult, IApiInterfaceImplementation):

    def __init__(self, internal_supervision_result: InternalSupervisionResult, /):
        super().__init__()

        self._context: Context | None = None

        self._template_path = internal_supervision_result.template.get_path()

        self._template = ExternalTemplate(internal_supervision_result.template)
        self._matching_result = ExternalMatchingResult(internal_supervision_result.matching_result, self._template)

        self._score = internal_supervision_result.score
        self._delta = internal_supervision_result.delta
        self._delta_prime = internal_supervision_result.delta_prime
        self._transformation_matrix = internal_supervision_result.transformation_matrix

        # noinspection PyProtectedMember
        self._match_weights: Dict[IMatch, float] = internal_supervision_result.get_match_weights()

    def set_api_context(self, context: Context, /) -> None:
        self._context = context

        # propagate the context further down the hierarchy of objects
        self._template.set_api_context(context)
        self._matching_result.set_api_context(context)

    def clear_api_context(self) -> None:
        self._context = None

        self._template.clear_api_context()
        self._matching_result.clear_api_context()

    @property
    def template(self) -> ExternalTemplate:
        return self._template

    @property
    def matching_result(self) -> ExternalMatchingResult:
        return self._matching_result

    @property
    def score(self) -> float:
        return self._score

    @property
    def delta(self) -> np.ndarray:
        return self._delta

    @property
    def delta_prime(self) -> np.ndarray:
        return self._delta_prime

    @property
    def transformation_matrix(self) -> np.ndarray:
        return self._transformation_matrix

    def interpret_async(self, /, *, target: IImage) -> Future:

        # TODO: this is hacky, maybe use a more clean approach here?
        assert isinstance(target, Image)

        assert self._context is not None, \
            ("The external superivision result has no context information, probably because it has been given to the API user "
             "before the context has been initialized in this object via the 'set_api_context' method, which is incorrect behavior.")

        _api_context = self._context

        self.clear_api_context()

        # noinspection PyProtectedMember
        return _api_context._submit_task(
            template_interpret,
            f"Interpreting [b]{self.template.name}[/]...",
            self._template_path,
            self,
            interpretation_target_path=target._path
        )

    def interpret(self, /, **kwargs) -> ExternalInterpretationResult:
        future = self.interpret_async(**kwargs)
        return future.result()

    def get_match_weight(self, match: IMatch, /) -> float:

        if match in self._match_weights:
            return self._match_weights[match]

        return 1.0



"""
Module represeting the OfficialEye context.
"""

from __future__ import annotations

from concurrent.futures import Future as PythonFuture
from concurrent.futures import ProcessPoolExecutor
from types import TracebackType
from typing import TYPE_CHECKING, Dict

from officialeye._api.future import Future
from officialeye._api.mutator import IMutator

# noinspection PyProtectedMember
from officialeye._api_builtins.init import initialize_builtins

# noinspection PyProtectedMember
from officialeye._internal.feedback.abstract import AbstractFeedbackInterface

# noinspection PyProtectedMember
from officialeye._internal.feedback.dummy import DummyFeedbackInterface
from officialeye.error.errors.general import ErrInvalidIdentifier
from officialeye.error.errors.internal import ErrInvalidState
from officialeye.error.errors.template import ErrTemplateInvalidMutator

if TYPE_CHECKING:
    from officialeye.types import ConfigDict, InterpretationFactory, MatcherFactory, MutatorFactory, SupervisorFactory


class Context:

    def __init__(self, /, *, afi: AbstractFeedbackInterface | None = None):
        self._entered: bool = False
        self._disposed: bool = False

        if afi is None:
            self._afi = DummyFeedbackInterface()
        else:
            self._afi = afi

        self._executor = ProcessPoolExecutor()

        self._mutator_factories: Dict[str, MutatorFactory] = {}
        self._matcher_factories: Dict[str, MatcherFactory] = {}
        self._supervisor_factories: Dict[str, SupervisorFactory] = {}
        self._interpretation_factories: Dict[str, InterpretationFactory] = {}

        # initialize with built-in mutators
        initialize_builtins(self)

    def _get_afi(self) -> AbstractFeedbackInterface:
        return self._afi

    def _submit_task(self, task, description: str, *args, **kwargs) -> Future:

        afi_fork = self._afi.fork(description)

        python_future: PythonFuture = self._executor.submit(
            task,
            *args,
            **kwargs,
            # Arguments that need to be always passed to the internal implementation when starting tasks.
            # It is very important that the argument dictionary is picklable, because it will be passed from the parent
            # process to a child process by the ProcessPoolExecutor.
            afi=afi_fork,
            mutator_factories=self._mutator_factories,
            matcher_factories=self._matcher_factories,
            supervisor_factories=self._supervisor_factories,
            interpretation_factories=self._interpretation_factories
        )

        return Future(self, python_future, afi_fork=afi_fork)

    def register_mutator(self, mutator_id: str, factory: MutatorFactory, /) -> None:

        if mutator_id in self._mutator_factories:
            raise ErrInvalidIdentifier(
                f"while adding the '{mutator_id}' mutator.",
                "A mutator with the same id has already been registered."
            )

        self._mutator_factories[mutator_id] = factory

    def register_matcher(self, matcher_id: str, factory: MatcherFactory, /) -> None:

        if matcher_id in self._matcher_factories:
            raise ErrInvalidIdentifier(
                f"while adding the '{matcher_id}' matcher.",
                "A matcher with the same id has already been registered."
            )

        self._matcher_factories[matcher_id] = factory

    def register_supervisor(self, supervisor_id: str, factory: SupervisorFactory, /) -> None:

        if supervisor_id in self._matcher_factories:
            raise ErrInvalidIdentifier(
                f"while adding the '{supervisor_id}' matcher.",
                "A supervisor with the same id has already been registered."
            )

        self._supervisor_factories[supervisor_id] = factory

    def register_interpretation(self, interpretation_id: str, factory: InterpretationFactory, /) -> None:

        if interpretation_id in self._interpretation_factories:
            raise ErrInvalidIdentifier(
                f"while adding the '{interpretation_id}' interpretation.",
                "An interpretation with the same id has already been registered."
            )

        self._interpretation_factories[interpretation_id] = factory

    def get_mutator(self, mutator_id: str, config: ConfigDict, /) -> IMutator:

        if mutator_id not in self._mutator_factories:
            raise ErrTemplateInvalidMutator(
                f"while looking for a factory generating mutator '{mutator_id}'.",
                "A mutator with this id has not been registered."
            )

        return self._mutator_factories[mutator_id](config)

    def __enter__(self):

        if self._entered:
            raise ErrInvalidState(
                "while entering the api context.",
                "The context has already been entered, which is illegal state."
            )

        self._entered = True
        return self

    def __exit__(self, exception_type: any, exception_value: BaseException | None, traceback: TracebackType | None):
        assert self._entered
        self._entered = False

        if self._disposed:
            raise ErrInvalidState(
                "while leaving the api context.",
                "The resources have already been disposed."
            )

        self.dispose(exception_type, exception_value, traceback)

    def dispose(self, exception_type: any = None, exception_value: BaseException | None = None, traceback: TracebackType | None = None) -> None:
        self._afi.dispose(exception_type, exception_value, traceback)
        self._executor.shutdown(wait=True)
        self._disposed = True



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    # noinspection PyProtectedMember
    from officialeye._api.context import Context


class IApiInterfaceImplementation(ABC):

    @abstractmethod
    def set_api_context(self, context: Context, /) -> None:
        """
        This method should be used to propagate the public API's context to the objects returned by the internal implementation of the API.
        Those objects are called 'external' and should be picklable if the API context has not yet been set via this method.
        If it was, then all methods guaranteed by the corresponding object's public API interface can be implemented properly.
        """
        raise NotImplementedError()

    @abstractmethod
    def clear_api_context(self) -> None:
        """
        This method should clear the public API's context stored in the current object, and in all internal objects implementing this interface.
        It is essential that after running this method, the object is picklable.
        """
        raise NotImplementedError()



"""
Module for abstracting out the ability to inject custom configurations specified using dictionaries.
The goal of this module is to provide a nice API for validated user-specified configurations
and safely retrieving information from there.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Callable

from officialeye.error.errors.general import ErrInvalidKey

if TYPE_CHECKING:
    from officialeye.types import ConfigDict


class Config(ABC):

    def __init__(self, config_dict: ConfigDict, /):
        self._config_dict = config_dict

    @abstractmethod
    def _get_invalid_key_error(self, key: str, /):
        raise NotImplementedError()

    def get(self, key: str, /, *, value_preprocessor: Callable[[str], any] | None = None, default=None):

        if key not in self._config_dict:

            if default is None:
                raise self._get_invalid_key_error(key)

            return default

        _value = self._config_dict[key]

        # apply value preprocessor
        if value_preprocessor is not None:
            _value = value_preprocessor(_value)

        return _value


class MutatorConfig(Config):

    def __init__(self, config_dict: ConfigDict, mutator_id: str, /):

        super().__init__(config_dict)

        self._mutator_id = mutator_id

    def _get_invalid_key_error(self, key: str, /):
        return ErrInvalidKey(
            f"while reading configuration of the '{self._mutator_id}' mutator.",
            f"Could not find a value for key '{key}'."
        )


class MatcherConfig(Config):

    def __init__(self, config_dict: ConfigDict, matcher_id: str, /):

        super().__init__(config_dict)

        self._matcher_id = matcher_id

    def _get_invalid_key_error(self, key: str, /):
        return ErrInvalidKey(
            f"while reading configuration of the '{self._matcher_id}' matcher.",
            f"Could not find a value for key '{key}'."
        )


class SupervisorConfig(Config):

    def __init__(self, config_dict: ConfigDict, matcher_id: str, /):

        super().__init__(config_dict)

        self._matcher_id = matcher_id

    def _get_invalid_key_error(self, key: str, /):
        return ErrInvalidKey(
            f"while reading configuration of the '{self._matcher_id}' supervisor.",
            f"Could not find a value for key '{key}'."
        )


class InterpretationConfig(Config):

    def __init__(self, config_dict: ConfigDict, interpretation_id: str, /):

        super().__init__(config_dict)

        self._interpretation_id = interpretation_id

    def _get_invalid_key_error(self, key: str, /):
        return ErrInvalidKey(
            f"while reading configuration of the '{self._interpretation_id}' interpretation.",
            f"Could not find a value for key '{key}'."
        )



from typing import Dict

# noinspection PyProtectedMember
from officialeye._api.mutator import IMutator
from officialeye._internal.context.singleton import get_internal_context


def load_mutator_from_dict(mutator_dict: Dict[str, any], /) -> IMutator:

    assert "id" in mutator_dict

    mutator_id = mutator_dict["id"]

    mutator_config = mutator_dict.get("config", {})

    return get_internal_context().get_mutator(mutator_id, mutator_config)



from __future__ import annotations

from typing import TYPE_CHECKING, Dict

import numpy as np

# noinspection PyProtectedMember
from officialeye._api.future import Future

# noinspection PyProtectedMember
from officialeye._api.template.interpretation_result import IInterpretationResult

# noinspection PyProtectedMember
from officialeye._api.template.supervision_result import ISupervisionResult
from officialeye.error.errors.general import ErrOperationNotSupported

if TYPE_CHECKING:
    # noinspection PyProtectedMember
    from officialeye._api.image import IImage

    # noinspection PyProtectedMember
    from officialeye._api.template.match import IMatch

    # noinspection PyProtectedMember
    from officialeye._api.template.supervision_result import SupervisionResult
    from officialeye._internal.template.internal_matching_result import InternalMatchingResult
    from officialeye._internal.template.internal_template import InternalTemplate


class InternalSupervisionResult(ISupervisionResult):

    def __init__(self, supervision_result: SupervisionResult, internal_template: InternalTemplate,
                 internal_matching_result: InternalMatchingResult, /):
        self._supervision_result = supervision_result
        self._internal_template = internal_template
        self._internal_matching_result = internal_matching_result

    @property
    def template(self) -> InternalTemplate:
        return self._internal_template

    @property
    def matching_result(self) -> InternalMatchingResult:
        return self._internal_matching_result

    @property
    def score(self) -> float:
        return self._supervision_result.get_score()

    @property
    def delta(self) -> np.ndarray:
        return self._supervision_result.delta

    @property
    def delta_prime(self) -> np.ndarray:
        return self._supervision_result.delta_prime

    @property
    def transformation_matrix(self) -> np.ndarray:
        return self._supervision_result.transformation_matrix

    def interpret_async(self, /, *, target: IImage) -> Future:
        raise ErrOperationNotSupported(
            "while accessing an internal supervision result instance.",
            "The way in which it was accessed is not supported."
        )

    def interpret(self, /, **kwargs) -> IInterpretationResult:
        raise ErrOperationNotSupported(
            "while accessing an internal supervision result instance.",
            "The way in which it was accessed is not supported."
        )

    def get_match_weights(self) -> Dict[IMatch, float]:
        # noinspection PyProtectedMember
        return self._supervision_result._match_weights

    def get_match_weight(self, match: IMatch, /) -> float:

        match_weights = self.get_match_weights()

        if match in match_weights:
            return match_weights[match]

        return 1.0



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Iterable

import numpy as np

from officialeye._api.template.region import IRegion

if TYPE_CHECKING:
    from officialeye._api.mutator import IMutator


class IFeature(IRegion, ABC):

    def __str__(self) -> str:
        return f"Feature '{self.identifier}'"

    @abstractmethod
    def get_mutators(self) -> Iterable[IMutator]:
        """
        Returns:
            A list of mutators from the feature class of the feature, in the order in which they are to be applied.
        """
        raise NotImplementedError()

    def apply_mutators_to_image(self, img: np.ndarray, /) -> np.ndarray:
        """
        Takes an image and applies the mutators defined in the corresponding feature class.

        Arguments:
            img: The image that should be transformed.

        Returns:
            The resulting image.
        """

        for mutator in self.get_mutators():
            img = mutator.mutate(img)

        return img



from officialeye._internal.context.context import InternalContext
from officialeye._internal.feedback.abstract import AbstractFeedbackInterface

_internal_context: InternalContext = InternalContext()


def get_internal_context() -> InternalContext:
    global _internal_context
    return _internal_context


def get_internal_afi() -> AbstractFeedbackInterface:
    return get_internal_context().get_afi()



from officialeye.error.codes import ERR_GENERAL
from officialeye.error.error import OEError
from officialeye.error.modules import ERR_MODULE_GENERAL


class ErrGeneral(OEError):

    def __init__(self, while_text: str, problem_text: str, /):
        super().__init__(ERR_MODULE_GENERAL, ERR_GENERAL[0], ERR_GENERAL[1], while_text, problem_text, is_regular=False)

        self._init_args = while_text, problem_text

    def __reduce__(self):
        return self.__class__, self._init_args


class ErrOperationNotSupported(ErrGeneral):

    def __init__(self, while_text: str, problem_text: str, /):
        super().__init__(while_text, problem_text)

        self._init_args = while_text, problem_text

    def __reduce__(self):
        return self.__class__, self._init_args


class ErrInvalidKey(ErrGeneral):

    def __init__(self, while_text: str, problem_text: str, /):
        super().__init__(while_text, problem_text)

        self._init_args = while_text, problem_text

    def __reduce__(self):
        return self.__class__, self._init_args


class ErrInvalidIdentifier(ErrGeneral):

    def __init__(self, while_text: str, problem_text: str, /):
        super().__init__(while_text, problem_text)

        self._init_args = while_text, problem_text

    def __reduce__(self):
        return self.__class__, self._init_args


class ErrObjectNotInitialized(ErrGeneral):

    def __init__(self, while_text: str, problem_text: str, /):
        super().__init__(while_text, problem_text)

        self._init_args = while_text, problem_text

    def __reduce__(self):
        return self.__class__, self._init_args


class ErrInvalidImage(ErrGeneral):

    def __init__(self, while_text: str, problem_text: str, /):
        super().__init__(while_text, problem_text)

        self._init_args = while_text, problem_text

    def __reduce__(self):
        return self.__class__, self._init_args



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List

# noinspection PyProtectedMember
from officialeye._api.template.match import IMatch
from officialeye._internal.context.singleton import get_internal_context
from officialeye._internal.template.shared_matching_result import SharedMatchingResult

if TYPE_CHECKING:
    from officialeye._internal.template.internal_template import InternalTemplate


class InternalMatchingResult(SharedMatchingResult):
    """
    Representation of the matching result, designed to be used by the child process only.
    """

    def __init__(self, template: InternalTemplate, /):
        super().__init__(template)

        self._template_id = template.identifier

        # keys: keypoint ids
        # values: matches with this keypoint
        self._matches_dict: Dict[str, List[IMatch]] = {}

        for keypoint in self.template.keypoints:
            self._matches_dict[keypoint.identifier] = []

    @property
    def template(self) -> InternalTemplate:
        return get_internal_context().get_template(self._template_id)



ERR_MODULE_GENERAL = "general"
ERR_MODULE_INTERNAL = "internal"
ERR_MODULE_IO = "io"
ERR_MODULE_MATCHING = "matching"
ERR_MODULE_SUPERVISION = "supervision"
ERR_MODULE_TEMPLATE = "template"

# TODO: remove the module system



import numpy as np

# noinspection PyProtectedMember
from officialeye._api.template.template import Template
from officialeye._cli.context import CLIContext
from officialeye._cli.utils import visualize_feature, visualize_keypoint

# noinspection PyProtectedMember
from officialeye._internal.feedback.verbosity import Verbosity


def _visualize_regions(template: Template, background_img: np.ndarray, /, *,
                       hide_features: bool = False, hide_keypoints: bool = False) -> np.ndarray:

    img = background_img

    if not hide_features:
        for feature in template.features:
            img = visualize_feature(feature, img)

    if not hide_keypoints:
        for keypoint in template.keypoints:
            img = visualize_keypoint(keypoint, img)

    return img


def do_show(context: CLIContext, /, *, template_path: str, **kwargs):

    # print OfficialEye logo and other introductory information (if necessary)
    context.print_intro()

    template = Template(context.get_api_context(), path=template_path)

    template_img = template.get_image().load()

    template_img_mutated = template.get_mutated_image().load()

    if template_img_mutated.shape == template_img.shape:
        background_img = template_img_mutated
    else:

        context.get_terminal_ui().warn(
            Verbosity.INFO_VERBOSE,
            f"One of the source mutators of the '{template.identifier}' template has changed the shape of the image. "
            f"To ensure that the regions of the template are visualized correctly, "
            f"the original template image had to be used as the background."
        )

        background_img = template_img

    visualization = _visualize_regions(template, background_img, **kwargs)

    context.export_and_show_image(visualization, file_name=f"{template.identifier}.png")



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, Iterable, List

# noinspection PyProtectedMember
from officialeye._api.future import Future

# noinspection PyProtectedMember
from officialeye._api.image import IImage, Image

# noinspection PyProtectedMember
from officialeye._api.mutator import IMutator

# noinspection PyProtectedMember
from officialeye._api.template.template_interface import ITemplate
from officialeye._internal.api.detect import template_detect
from officialeye._internal.api_implementation import IApiInterfaceImplementation

# noinspection PyProtectedMember
from officialeye._internal.template.external_feature import ExternalFeature
from officialeye._internal.template.keypoint import ExternalKeypoint
from officialeye.error.errors.general import ErrOperationNotSupported

if TYPE_CHECKING:
    # noinspection PyProtectedMember
    from officialeye._api.context import Context

    # noinspection PyProtectedMember
    from officialeye._api.template.supervision_result import ISupervisionResult
    from officialeye._internal.template.internal_template import InternalTemplate


class ExternalTemplate(ITemplate, IApiInterfaceImplementation):
    """
    Representation of a template instance designed to be shared between processes.
    It is very important that this class is picklable!
    """

    def __init__(self, template: InternalTemplate, /):
        super().__init__()

        self._context: Context | None = None

        self._identifier: str = template.identifier
        self._name: str = template.name
        self._path: str = template.get_path()
        self._source_image_path: str = template.get_source_image_path()

        self._width = template.width
        self._height = template.height

        self._keypoints: Dict[str, ExternalKeypoint] = {}
        self._features: Dict[str, ExternalFeature] = {}

        for keypoint in template.keypoints:
            self._keypoints[keypoint.identifier] = ExternalKeypoint(keypoint, self)

        for feature in template.features:
            self._features[feature.identifier] = ExternalFeature(feature, self)

        self._source_mutators: List[IMutator] = [
            mutator for mutator in template.get_source_mutators()
        ]

        self._target_mutators: List[IMutator] = [
            mutator for mutator in template.get_target_mutators()
        ]

    def set_api_context(self, context: Context, /) -> None:
        self._context = context

        for external_keypoint in self.keypoints:
            external_keypoint.set_api_context(context)

        for external_feature in self.features:
            external_feature.set_api_context(context)

    def clear_api_context(self) -> None:
        self._context = None

        for external_keypoint in self.keypoints:
            external_keypoint.clear_api_context()

        for external_feature in self.features:
            external_feature.clear_api_context()

    def load(self) -> None:
        raise ErrOperationNotSupported(
            "while accessing an external template instance.",
            "The way in which it was accessed is not supported."
        )

    def detect_async(self, /, *, target: IImage) -> Future:

        # TODO: this is hacky, maybe use a more clean approach here?
        assert isinstance(target, Image)

        # noinspection PyProtectedMember
        return self._context._submit_task(
            template_detect,
            f"Detecting [b]{self._name}[/]...",
            self._path,
            target_path=target._path,
        )

    def detect(self, /, **kwargs) -> ISupervisionResult:
        future = self.detect_async(**kwargs)
        return future.result()

    def get_image(self) -> IImage:
        return Image(self._context, path=self._source_image_path)

    def get_mutated_image(self) -> IImage:
        img = self.get_image()
        img.apply_mutators(*self._source_mutators)
        return img

    @property
    def identifier(self) -> str:
        return self._identifier

    @property
    def name(self) -> str:
        return self._name

    @property
    def width(self) -> int:
        return self._width

    @property
    def height(self) -> int:
        return self._height

    @property
    def keypoints(self) -> Iterable[ExternalKeypoint]:
        return self._keypoints.values()

    @property
    def features(self) -> Iterable[ExternalFeature]:
        return self._features.values()

    def get_feature(self, feature_id: str, /) -> ExternalFeature | None:
        if feature_id not in self._features:
            return None
        return self._features[feature_id]

    def get_keypoint(self, keypoint_id: str, /) -> ExternalKeypoint | None:
        if keypoint_id not in self._keypoints:
            return None
        return self._keypoints[keypoint_id]



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

import numpy as np

from officialeye._api.config import MutatorConfig

if TYPE_CHECKING:
    from officialeye.types import ConfigDict


class IMutator(ABC):

    @property
    def config(self) -> MutatorConfig:
        raise NotImplementedError()

    @abstractmethod
    def mutate(self, img: np.ndarray, /) -> np.ndarray:
        raise NotImplementedError()


class Mutator(IMutator, ABC):

    def __init__(self, mutator_id: str, config_dict: ConfigDict, /):
        super().__init__()

        self.mutator_id = mutator_id

        self._config = MutatorConfig(config_dict, mutator_id)

    @property
    def config(self) -> MutatorConfig:
        return self._config



import os
import random
from tempfile import NamedTemporaryFile
from types import TracebackType
from typing import List

import click
import cv2
import numpy as np
from rich.prompt import Confirm

from officialeye import Context
from officialeye.__version__ import __ascii_logo__
from officialeye._cli.ui import TerminalUI, Verbosity


class CLIContext:

    def __init__(self, **kwargs):
        self._api: Context | None = None
        self._ui: TerminalUI | None = None

        self.handle_exceptions = True
        self.visualization_generation = False
        self.export_directory = None

        self.verbosity = Verbosity.QUIET
        self.disable_logo = False

        self._export_counter = 1
        self._not_deleted_temporary_files: List[str] = []

        self.set_params(**kwargs)

    def set_params(self, /, *, handle_exceptions: bool | None = None, visualization_generation: bool | None = None,
                   export_directory: str | None = None, verbosity: Verbosity | None = None, disable_logo: bool | None = None):
        if handle_exceptions is not None:
            self.handle_exceptions = handle_exceptions

        if visualization_generation is not None:
            self.visualization_generation = visualization_generation
        if export_directory is not None:
            self.export_directory = export_directory

        if verbosity is not None:
            self.verbosity = verbosity
        if disable_logo is not None:
            self.disable_logo = disable_logo

    def __enter__(self):
        assert self._api is None
        assert self._ui is None

        assert self._export_counter == 1
        assert len(self._not_deleted_temporary_files) == 0

        self._ui = TerminalUI(self.verbosity)
        self._api = Context(afi=self._ui)

        return self

    def dispose(self):

        self._assert_entered()

        if len(self._not_deleted_temporary_files) > 0 and (self.verbosity == Verbosity.QUIET or Confirm.ask(
                ":question: Should temporary files created above be cleaned up now?",
                default=True, console=self._ui.get_console()
            )
        ):
            files_removed = 0

            # cleanup temporary files
            for temp_file_path in self._not_deleted_temporary_files:
                if os.path.isfile(temp_file_path):
                    os.unlink(temp_file_path)
                    files_removed += 1

            self._ui.info(Verbosity.INFO, f"Successfully removed {files_removed} temporary file(s).")

        # reset fields related to file exporting
        self._export_counter = 1
        self._not_deleted_temporary_files = []

        # dispose the api together with all resources it manages, such as the AFI
        self._api.dispose()

        self._api = None
        self._ui = None

    def __exit__(self, exception_type: any, exception_value: BaseException, traceback: TracebackType):

        self._assert_entered()

        if not self.handle_exceptions:
            self.dispose()
            # tell python that we do not want to handle the exception
            return None

        # handle the possible exception
        if exception_value is None:
            self.dispose()
            return None

        self._ui.handle_uncaught_error(exception_type, exception_value, traceback)

        # free all allocated resources
        self.dispose()

        # tell python that we have handled the exception ourselves
        return True

    def print_logo(self):

        if self.disable_logo:
            return

        logo_color = random.choice([
            "purple",
            "yellow",
            "red",
            "green",
            "cyan"
        ])

        self._ui.echo(Verbosity.INFO, __ascii_logo__, style=f"bold {logo_color}")

    def print_intro(self):

        self.print_logo()

        # print preliminary warnings if necessary
        if not self.handle_exceptions:
            self._ui.warn(Verbosity.INFO, "Raw error mode enabled. Use this mode only if you know precisely what you are doing!", )

        if self.verbosity >= Verbosity.DEBUG:
            self._ui.warn(Verbosity.INFO, "Debug mode enabled. Disable for production use to improve performance.")

        if self.visualization_generation:
            self._ui.warn(Verbosity.INFO, "Visualization generation mode enabled. Disable for production use to improve performance.")

    def _assert_entered(self):
        assert self._api is not None, "The context must be entered for this method to work correctly."
        assert self._ui is not None, "The context must be entered for this method to work correctly."

    def get_api_context(self) -> Context:
        self._assert_entered()
        return self._api

    def get_terminal_ui(self) -> TerminalUI:
        self._assert_entered()
        return self._ui

    def _allocate_file_name(self) -> str:
        self._assert_entered()
        file_name = "%03d.png" % self._export_counter
        self._export_counter += 1
        return file_name

    def allocate_file_for_export(self, /, *, file_name: str = "") -> str:

        self._assert_entered()

        file_suffix = ".png" if file_name == "" else file_name

        if self.export_directory is None:
            with NamedTemporaryFile(prefix="officialeye_", suffix=f"_{file_suffix}", delete=False) as fp:
                fp.close()
            self._not_deleted_temporary_files.append(fp.name)
            return fp.name

        if file_name == "":
            file_name = self._allocate_file_name()

        return os.path.join(self.export_directory, file_name)

    def export_image(self, img: np.ndarray, /, *, file_name: str = "") -> str:

        export_file_path = self.allocate_file_for_export(file_name=file_name)

        cv2.imwrite(export_file_path, img)

        self._ui.info(Verbosity.INFO, f"Exported [b]{export_file_path}[/].")

        return export_file_path

    def export_and_show_image(self, img: np.ndarray, /, *, file_name: str = ""):
        path = self.export_image(img, file_name=file_name)

        if self.verbosity != Verbosity.QUIET and Confirm.ask(
            ":question: Would you like to open the image in an image viewer (as provided by the OS)?",
            default=True, console=self._ui.get_console()
        ):
            click.launch(path, locate=False)



from __future__ import annotations

from typing import TYPE_CHECKING

# noinspection PyProtectedMember
from officialeye._api.template.keypoint import IKeypoint
from officialeye._internal.api_implementation import IApiInterfaceImplementation
from officialeye._internal.template.region import ExternalRegion, InternalRegion
from officialeye.error.errors.template import ErrTemplateInvalidKeypoint

if TYPE_CHECKING:
    # noinspection PyProtectedMember
    from officialeye._api.context import Context
    from officialeye._internal.template.external_template import ExternalTemplate


class InternalKeypoint(InternalRegion, IKeypoint):

    def __init__(self, template_id: str, keypoint_dict: dict, /):
        super().__init__(template_id, keypoint_dict)

        self._matches_min = keypoint_dict["matches"]["min"]
        self._matches_max = keypoint_dict["matches"]["max"]

        if self._matches_max < self._matches_min:
            raise ErrTemplateInvalidKeypoint(
                f"while loading template keypoint '{self.identifier}'",
                f"the lower bound on the match count ({self._matches_min}) exceeds the upper bound ({self._matches_max})"
            )

        if self._matches_min < 0:
            raise ErrTemplateInvalidKeypoint(
                f"while loading template keypoint '{self.identifier}'",
                f"the lower bound on the match count ({self._matches_min}) cannot be negative"
            )

        assert 0 <= self._matches_min <= self._matches_max

    @property
    def matches_min(self) -> int:
        return self._matches_min

    @property
    def matches_max(self) -> int:
        return self._matches_max


class ExternalKeypoint(ExternalRegion, IKeypoint, IApiInterfaceImplementation):

    def __init__(self, internal_keypoint: InternalKeypoint, external_template: ExternalTemplate, /):
        super().__init__(internal_keypoint, external_template)

        self._matches_min = internal_keypoint.matches_min
        self._matches_max = internal_keypoint.matches_max

    @property
    def matches_min(self) -> int:
        return self._matches_min

    @property
    def matches_max(self) -> int:
        return self._matches_max

    def set_api_context(self, context: Context, /) -> None:
        # no methods of this class require any contextual information to work, nothing to do
        pass

    def clear_api_context(self) -> None:
        pass



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Iterable

from officialeye._api.template.match import IMatch

if TYPE_CHECKING:
    from officialeye._api.template.template_interface import ITemplate


class IMatchingResult(ABC):

    @property
    @abstractmethod
    def template(self) -> ITemplate:
        raise NotImplementedError()

    @abstractmethod
    def get_all_matches(self) -> Iterable[IMatch]:
        raise NotImplementedError()

    @abstractmethod
    def get_total_match_count(self) -> int:
        raise NotImplementedError()

    @abstractmethod
    def get_matches_for_keypoint(self, keypoint_id: str, /) -> Iterable[IMatch]:
        raise NotImplementedError()



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from pytesseract import pytesseract

# noinspection PyProtectedMember
from officialeye._api.template.interpretation import Interpretation

if TYPE_CHECKING:
    # noinspection PyProtectedMember
    from officialeye._api.template.feature import IFeature
    from officialeye.types import ConfigDict, FeatureInterpretation


class TesseractInterpretation(Interpretation):

    INTERPRETATION_ID = "ocr_tesseract"

    def __init__(self, config_dict: ConfigDict, /):
        super().__init__(TesseractInterpretation.INTERPRETATION_ID, config_dict)

        self._tesseract_lang = self.config.get("lang", default="eng", value_preprocessor=str)
        self._tesseract_config = self.config.get("config", default="", value_preprocessor=str)

    def interpret(self, feature_img: np.ndarray, feature: IFeature, /) -> FeatureInterpretation:
        return pytesseract.image_to_string(feature_img, lang=self._tesseract_lang, config=self._tesseract_config).strip()



import time
from typing import Tuple


class Timer:

    def __init__(self):
        self._start_time: Tuple[float, float] | None = None
        self._end_time: Tuple[float, float] | None = None

    def __enter__(self):
        self._start_time = time.perf_counter(), time.process_time()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._end_time = time.perf_counter(), time.process_time()

    def get_real_time(self) -> float:
        assert self._start_time is not None
        assert self._end_time is not None
        return self._end_time[0] - self._start_time[0]

    def get_cpu_time(self) -> float:
        assert self._start_time is not None
        assert self._end_time is not None
        return self._end_time[1] - self._start_time[1]



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

import numpy as np

from officialeye._api.config import InterpretationConfig
from officialeye._api.template.feature import IFeature

if TYPE_CHECKING:
    from officialeye.types import ConfigDict, FeatureInterpretation


class IInterpretation(ABC):

    @property
    def config(self) -> InterpretationConfig:
        raise NotImplementedError()

    @abstractmethod
    def interpret(self, feature_img: np.ndarray, feature: IFeature, /) -> FeatureInterpretation:
        raise NotImplementedError()


class Interpretation(IInterpretation, ABC):

    def __init__(self, interpretation_id: str, config_dict: ConfigDict, /):
        super().__init__()

        self.interpretation_id = interpretation_id

        self._config = InterpretationConfig(config_dict, interpretation_id)

    @property
    def config(self) -> InterpretationConfig:
        return self._config



"""
Module containing the OfficialEye internal API.

WARNING!
Do not import it unless you know precisely what you are doing.
Instead, use the public API to interact with OfficialEye programatically.
"""

import z3

z3.set_param("parallel.enable", True)



"""
This module contains everything that should be visible to the API user at the top level of imports,
i.e., all symbols that should be importable via `from officialeye import symbol`
"""

# TODO: write docstrings for all public methods of the api



import os
from typing import List

import cv2
import numpy as np

# noinspection PyProtectedMember
from officialeye._api.image import IImage

# noinspection PyProtectedMember
from officialeye._api.mutator import IMutator
from officialeye._internal.context.singleton import get_internal_afi
from officialeye._internal.feedback.verbosity import Verbosity
from officialeye.error.errors.io import ErrIOInvalidPath


class InternalImage(IImage):

    def __init__(self, /, *, path: str):
        super().__init__()

        self._mutators: List[IMutator] = []
        self._path = path

    def load(self) -> np.ndarray:

        if not os.path.isfile(self._path):
            raise ErrIOInvalidPath(
                f"while loading image located at '{self._path}'.",
                "This path does not refer to a file."
            )

        if not os.access(self._path, os.R_OK):
            raise ErrIOInvalidPath(
                f"while loading image located at '{self._path}'.",
                "The file at this path is not readable."
            )

        img = cv2.imread(self._path, cv2.IMREAD_COLOR)

        for mutator in self._mutators:
            get_internal_afi().info(Verbosity.DEBUG, f"InternalImage::load() applies mutator '{mutator}'")
            img = mutator.mutate(img)

        return img

    def apply_mutators(self, *mutators: IMutator):
        self._mutators += mutators



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Dict, TypeAlias, Union

# noinspection PyProtectedMember
from officialeye._api.mutator import IMutator

# noinspection PyProtectedMember
from officialeye._api.template.interpretation import IInterpretation

# noinspection PyProtectedMember
from officialeye._api.template.matcher import IMatcher

# noinspection PyProtectedMember
from officialeye._api.template.supervisor import ISupervisor

if TYPE_CHECKING:
    ConfigValue = Union[str, int, float, Dict[str, "ConfigValue"]]
    ConfigDict = Dict[str, ConfigValue]

    MutatorFactory = Callable[[ConfigDict], IMutator]
    MatcherFactory = Callable[[ConfigDict], IMatcher]
    SupervisorFactory = Callable[[ConfigDict], ISupervisor]
    InterpretationFactory = Callable[[ConfigDict], IInterpretation]

    FeatureInterpretation: TypeAlias = dict[str, "FeatureInterpretation"] | list["FeatureInterpretation"] | str | int | float | bool | None



from __future__ import annotations

from typing import TYPE_CHECKING, Any, TypeVar

if TYPE_CHECKING:
    from typing import Protocol

    SpinnerT = TypeVar("SpinnerT", bound="Spinner")

    class Spinner(Protocol):
        def update(self, text: str) -> None:
            ...

        def __enter__(self: SpinnerT) -> SpinnerT:
            ...

        def __exit__(self, *args: Any) -> None:
            ...

    class RichProtocol(Protocol):
        def __rich__(self) -> str:
            ...



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from officialeye._api.template.feature import IFeature

if TYPE_CHECKING:
    from officialeye._api.template.template_interface import ITemplate
    from officialeye.types import FeatureInterpretation


class IInterpretationResult(ABC):

    @property
    @abstractmethod
    def template(self) -> ITemplate:
        raise NotImplementedError()

    @abstractmethod
    def get_feature_interpretation(self, feature: IFeature, /) -> FeatureInterpretation:
        raise NotImplementedError()



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

import numpy as np

from officialeye._api.template.keypoint import IKeypoint

if TYPE_CHECKING:
    from officialeye._api.template.template import ITemplate


class IMatch(ABC):

    @property
    @abstractmethod
    def template(self) -> ITemplate:
        raise NotImplementedError()

    @property
    @abstractmethod
    def keypoint(self) -> IKeypoint:
        raise NotImplementedError()

    @property
    @abstractmethod
    def keypoint_point(self) -> np.ndarray:
        raise NotImplementedError()

    @property
    @abstractmethod
    def target_point(self) -> np.ndarray:
        raise NotImplementedError()

    @abstractmethod
    def get_score(self) -> float:
        raise NotImplementedError()

    @property
    def template_point(self) -> np.ndarray:
        """Returns the coordinates of the point lying in the keypoint, in the coordinate system of the underlying template."""
        return self.keypoint_point + self.keypoint.top_left

    def __str__(self) -> str:
        return (f"Point ({self.target_point[0]}, {self.target_point[1]}) matches ({self.keypoint_point[0]}, {self.keypoint_point[1]}) "
                f"in '{self.keypoint}' of '{self.template.identifier}'.")

    def __eq__(self, o: Any) -> bool:

        if not isinstance(o, IMatch):
            return False

        if self.template != o.template:
            return False

        if self.keypoint != o.keypoint:
            return False

        return (np.array_equal(self.keypoint_point, o.keypoint_point)
                and np.array_equal(self.target_point, o.target_point))

    def __lt__(self, o: Any) -> bool:
        assert isinstance(o, Match)
        return self.get_score() < o.get_score()

    def __hash__(self):
        return hash((
            self.template.identifier,
            self.keypoint.identifier,
            np.dot(self.keypoint_point, self.keypoint_point),
            np.dot(self.target_point, self.target_point)
        ))


class Match(IMatch):

    # TODO: remove the first argument from the constructor, and possibly create InternalMatch and ExternalMatch instances
    def __init__(self, template: ITemplate, keypoint: IKeypoint, /, *,
                 keypoint_point: np.ndarray, target_point: np.ndarray, score: float = 0.0):
        super().__init__()

        self._template = template
        self._keypoint = keypoint

        self._keypoint_point = keypoint_point
        self._target_point = target_point

        self._score = score

    def get_score(self) -> float:
        return self._score

    def set_score(self, new_score: float, /):
        self._score = new_score

    @property
    def template(self) -> ITemplate:
        return self._template

    @property
    def keypoint(self) -> IKeypoint:
        return self._keypoint

    @property
    def keypoint_point(self) -> np.ndarray:
        return self._keypoint_point.copy()

    @property
    def target_point(self) -> np.ndarray:
        return self._target_point.copy()



from typing import List

import numpy as np

# noinspection PyProtectedMember
from officialeye._api.detection import detect

# noinspection PyProtectedMember
from officialeye._api.image import Image

# noinspection PyProtectedMember
from officialeye._api.template.template import Template

# noinspection PyProtectedMember
from officialeye._api.template.template_interface import ITemplate
from officialeye._cli.context import CLIContext
from officialeye._cli.utils import visualize_feature

# noinspection PyProtectedMember
from officialeye._internal.feedback.verbosity import Verbosity


def _get_background(context: CLIContext, template: ITemplate, /) -> np.ndarray:

    raw_image = template.get_image().load()
    mutated_image = template.get_mutated_image().load()

    if raw_image.shape == mutated_image.shape:
        return mutated_image

    context.get_terminal_ui().warn(
        Verbosity.INFO,
        f"Could not use the mutated version of the '{template.identifier}' template "
        f"because one of the source mutators did not preserve the shape of the image. "
        f"Falling back to the non-mutated version of the source image."
    )

    return raw_image


def do_test(context: CLIContext, /, *,
            target_path: str, template_paths: List[str], show_features: bool):
    # print OfficialEye logo and other introductory information (if necessary)
    context.print_intro()

    api_context = context.get_api_context()

    target_image = Image(api_context, path=target_path)

    templates = [Template(api_context, path=template_path) for template_path in template_paths]

    result = detect(api_context, *templates, target=target_image)

    visualization = _get_background(context, result.template)
    target_image_mat = target_image.load()

    for feature in result.template.features:
        feature_image_mat = result.warp_feature(feature, target_image_mat)

        assert feature_image_mat.shape == (feature.h, feature.w, 3)

        feature_image_mutated_mat = feature.apply_mutators_to_image(feature_image_mat)

        if feature_image_mat.shape == feature_image_mutated_mat.shape:
            # mutators didn't change the shape of the image
            feature.insert_into_image(visualization, feature_image_mutated_mat)
        else:
            # some mutator has altered the shape of the feature image.
            # this means that we can no longer safely insert the mutated feature into the visualization.
            # therefore, we have to fall back to inserting the feature image unmutated
            context.get_terminal_ui().warn(
                Verbosity.INFO,
                f"Could not visualize the '{feature.identifier}' feature of the '{feature.template.identifier}' template, "
                f"because one of the mutators (corresponding to this feature) did not preserve the shape of the image. "
                f"Falling back to the non-mutated version of the feature image."
            )

            feature.insert_into_image(visualization, feature_image_mat)

    if show_features:
        # visualize features on the image
        for feature in result.template.features:
            visualization = visualize_feature(feature, visualization)

    context.export_and_show_image(visualization, file_name=f"{result.template.identifier}.png")



from __future__ import annotations

from concurrent.futures import ALL_COMPLETED
from concurrent.futures import Future as PythonFuture
from concurrent.futures import wait as python_wait
from typing import TYPE_CHECKING, Any, Dict, Iterable, Set, Tuple

# noinspection PyProtectedMember
from officialeye._internal.api_implementation import IApiInterfaceImplementation

# noinspection PyProtectedMember
from officialeye._internal.feedback.abstract import AbstractFeedbackInterface

if TYPE_CHECKING:
    from officialeye._api.context import Context


class Future:

    def __init__(self, context: Context, python_future: PythonFuture, /, *, afi_fork: AbstractFeedbackInterface):
        self._context = context
        self._future = python_future
        self._afi_fork = afi_fork

        self._afi_joined = False

    def cancel(self) -> bool:
        """
        Attempt to cancel the call.
        If the call is currently being executed and cannot be canceled, then the method will return False,
        otherwise the call will be canceled, and the method will return True.
        """
        return self._future.cancel()

    def cancelled(self) -> bool:
        """ Return True if the call was successfully canceled. """
        return self._future.cancelled()

    def running(self) -> bool:
        """ Return True if the call is currently being executed and cannot be canceled. """
        return self._future.running()

    def done(self) -> bool:
        """ Return True if the call was successfully canceled or finished running. """
        return self._future.done()

    def _afi_join(self):
        if not self._afi_joined:
            self._afi_joined = True
            # noinspection PyProtectedMember
            self._context._get_afi().join(self._afi_fork, self._future)

    def result(self, timeout: float | None = None) -> Any:
        """
        Return the value returned by the call. If the call hasn’t yet completed, then this method will wait up to timeout seconds.
        If the call hasn’t completed in timeout seconds, then a TimeoutError will be raised. Timeout can be an int or float.
        If timeout is not specified or None, there is no limit to the wait time.

        If the future is canceled before completing, then CancelledError will be raised.

        If the call raised, this method will raise the same exception.
        """

        result = self._future.result(timeout=timeout)

        assert isinstance(result, IApiInterfaceImplementation), \
            "Every call to an internal API function should return a proper public API interface implementation"

        result.set_api_context(self._context)

        self._afi_join()

        return result

    def exception(self, timeout: float | None = None) -> Any:
        """
        Return the exception raised by the call.
        If the call hasn’t yet completed, then this method will wait up to timeout seconds.
        If the call hasn’t completed in timeout seconds, then a TimeoutError will be raised.
        Timeout can be an int or float.
        If timeout is not specified or None, there is no limit to the wait time.

        If the future is canceled before completing, then CancelledError will be raised.

        If the call completed without raising, None is returned.
        """

        err = self._future.exception(timeout=timeout)

        self._afi_join()

        return err


def wait(futures: Iterable[Future], /, *, timeout: float | None = None, return_when=ALL_COMPLETED) -> Tuple[Set[Future], Set[Future]]:

    # noinspection PyProtectedMember
    python_futures = (f._future for f in futures)

    original_futures: Dict[PythonFuture, Future] = {}

    for future in futures:
        # noinspection PyProtectedMember
        original_futures[future._future] = future

    done, not_done = python_wait(python_futures, timeout=timeout, return_when=return_when)

    corresponding_done = set((original_futures[d] for d in done))
    corresponding_not_done = set((original_futures[d] for d in not_done))

    return corresponding_done, corresponding_not_done



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Iterable

from officialeye._api.config import SupervisorConfig
from officialeye._api.template.matching_result import IMatchingResult
from officialeye._api.template.supervision_result import SupervisionResult
from officialeye._api.template.template_interface import ITemplate

if TYPE_CHECKING:
    from officialeye.types import ConfigDict


class ISupervisor(ABC):

    @property
    @abstractmethod
    def config(self) -> SupervisorConfig:
        raise NotImplementedError()

    @abstractmethod
    def setup(self, template: ITemplate, matching_result: IMatchingResult, /) -> None:
        raise NotImplementedError()

    @abstractmethod
    def supervise(self, template: ITemplate, matching_result: IMatchingResult, /) -> Iterable[SupervisionResult]:
        raise NotImplementedError()


class Supervisor(ISupervisor, ABC):

    def __init__(self, supervisor_id: str, config_dict: ConfigDict, /):
        super().__init__()

        self._supervisor_id = supervisor_id
        self._config = SupervisorConfig(config_dict, supervisor_id)

    @property
    def config(self) -> SupervisorConfig:
        return self._config



from __future__ import annotations

import os
from typing import TYPE_CHECKING

import cv2
import numpy as np

# noinspection PyProtectedMember
from officialeye._api.template.interpretation import Interpretation

if TYPE_CHECKING:
    # noinspection PyProtectedMember
    from officialeye._api.template.feature import IFeature
    from officialeye.types import ConfigDict, FeatureInterpretation


class FileInterpretation(Interpretation):

    INTERPRETATION_ID = "file"

    def __init__(self, config_dict: ConfigDict, /):
        super().__init__(FileInterpretation.INTERPRETATION_ID, config_dict)

        self._path = self.config.get("path", value_preprocessor=str)

    def interpret(self, feature_img: np.ndarray, feature: IFeature, /) -> FeatureInterpretation:

        os.makedirs(os.path.dirname(self._path), exist_ok=True)

        cv2.imwrite(self._path, feature_img)

        return None



from __future__ import annotations

import sys
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Dict

import cv2
import numpy as np

from officialeye._api.template.feature import IFeature
from officialeye._api.template.interpretation_result import IInterpretationResult
from officialeye._api.template.match import IMatch
from officialeye.error.errors.general import ErrObjectNotInitialized

if TYPE_CHECKING:
    from officialeye._api.future import Future
    from officialeye._api.image import IImage
    from officialeye._api.template.matching_result import IMatchingResult
    from officialeye._api.template.template_interface import ITemplate


class ISupervisionResult(ABC):

    @property
    @abstractmethod
    def template(self) -> ITemplate:
        raise NotImplementedError()

    @property
    @abstractmethod
    def matching_result(self) -> IMatchingResult:
        raise NotImplementedError()

    @property
    @abstractmethod
    def score(self) -> float:
        raise NotImplementedError()

    @property
    @abstractmethod
    def delta(self) -> np.ndarray:
        raise NotImplementedError()

    @property
    @abstractmethod
    def delta_prime(self) -> np.ndarray:
        raise NotImplementedError()

    @property
    @abstractmethod
    def transformation_matrix(self) -> np.ndarray:
        raise NotImplementedError()

    def translate(self, template_point: np.ndarray, /) -> np.ndarray:
        """
        Translates the given template point into a target point. That is, given a position in the template's coordinate system, this function
        outputs the corresponding position in the target image's coordinate system, according to the affine transformation model.
        """
        assert template_point.shape == (2,)
        return self.transformation_matrix @ (template_point - self.delta) + self.delta_prime

    @abstractmethod
    def get_match_weight(self, match: IMatch, /) -> float:
        raise NotImplementedError()

    @abstractmethod
    def interpret_async(self, /, *, target: IImage) -> Future:
        raise NotImplementedError()

    @abstractmethod
    def interpret(self, /, **kwargs) -> IInterpretationResult:
        raise NotImplementedError()

    def get_weighted_mse(self, /) -> float:

        error = 0.0
        singificant_match_count = 0

        for match in self.matching_result.get_all_matches():

            match_weight = self.get_match_weight(match)

            if match_weight < sys.float_info.epsilon:
                continue

            singificant_match_count += 1

            s = match.template_point

            # calculate prediction
            p = self.translate(s)

            # calculate destination
            d = match.target_point

            current_error = p - d
            current_error_value = np.dot(current_error, current_error)

            error += current_error_value * match_weight

        return error / singificant_match_count

    def warp_feature(self, feature: IFeature, target: np.ndarray, /) -> np.ndarray:

        target_tl = self.translate(feature.top_left)
        target_tr = self.translate(feature.top_right)
        target_bl = self.translate(feature.bottom_left)
        target_br = self.translate(feature.bottom_right)

        dest_tl = np.array([0, 0], dtype=np.float64)
        dest_tr = np.array([feature.w, 0], dtype=np.float64)
        dest_br = np.array([feature.w, feature.h], dtype=np.float64)
        dest_bl = np.array([0, feature.h], dtype=np.float64)

        source_points = [target_tl, target_tr, target_br, target_bl]
        destination_points = [dest_tl, dest_tr, dest_br, dest_bl]

        homography = cv2.getPerspectiveTransform(np.float32(source_points), np.float32(destination_points))

        return cv2.warpPerspective(
            target,
            np.float32(homography),
            (feature.w, feature.h),
            flags=cv2.INTER_LINEAR
        )


class SupervisionResult:

    def __init__(self, /, **kwargs):
        # offset in the template's coordinates
        self._delta: np.ndarray | None = None
        # offset in the target image's coordinates
        self._delta_prime: np.ndarray | None = None

        self._transformation_matrix: np.ndarray | None = None

        # keys: matches
        # values: weights assigned by the supervision engine to each match (assigning is optional)
        # the higher the weight, the more we trust the correctness of the match and the greater its individual impact should be.
        # by default, the weight is 1.
        self._match_weights: Dict[IMatch, float] = {}

        # an optional value the supervision engine can set, representing how confident the engine is in the result
        self._score = 0.0

        self.set(**kwargs)

    def set(self, /, *, delta: np.ndarray | None = None, delta_prime: np.ndarray | None = None,
            transformation_matrix: np.ndarray | None = None, score: float | None = None):

        if delta is not None:
            assert delta.shape == (2,)
            self._delta = delta

        if delta_prime is not None:
            assert delta_prime.shape == (2,)
            self._delta_prime = delta_prime

        if transformation_matrix is not None:
            assert transformation_matrix.shape == (2, 2)
            self._transformation_matrix = transformation_matrix

        if score is not None:
            self._score = score

    def set_match_weight(self, match: IMatch, weight: float, /):
        assert weight >= 0
        self._match_weights[match] = weight

    def get_score(self) -> float:
        assert self._score >= 0.0
        return self._score

    @property
    def delta(self) -> np.ndarray:

        if self._delta is None:
            raise ErrObjectNotInitialized(
                "while trying to access the 'delta' parameter of the supervision result instance.",
                "This parameter has not been set."
            )

        return self._delta.copy()

    @property
    def delta_prime(self) -> np.ndarray:

        if self._delta_prime is None:
            raise ErrObjectNotInitialized(
                "while trying to access the 'delta_prime' parameter of the supervision result instance.",
                "This parameter has not been set."
            )

        return self._delta_prime.copy()

    @property
    def transformation_matrix(self) -> np.ndarray:

        if self._transformation_matrix is None:
            raise ErrObjectNotInitialized(
                "while trying to access the 'transformation_matrix' parameter of the supervision result instance.",
                "This parameter has not been set."
            )

        return self._transformation_matrix.copy()



from __future__ import annotations

from typing import TYPE_CHECKING

import cv2

from officialeye._internal.context.singleton import get_internal_context

# noinspection PyProtectedMember
from officialeye._internal.template.external_interpretation_result import ExternalInterpretationResult
from officialeye._internal.template.schema.loader import load_template

if TYPE_CHECKING:
    # noinspection PyProtectedMember
    from officialeye._api.template.supervision_result import ISupervisionResult


def template_interpret(template_path: str, supervision_result: ISupervisionResult, /, *,
                       interpretation_target_path: str, **kwargs) -> ExternalInterpretationResult:

    with get_internal_context().setup(**kwargs):

        template = load_template(template_path)

        interpretation_target = cv2.imread(interpretation_target_path, cv2.IMREAD_COLOR)

        # TODO: make sure that the target image and the interpretation target images have the same shape, similar to the following snippet
        """
            if target.shape != interpretation_target.shape:
                raise ErrInvalidImage(
                    "while making sure that the target image and the interpretation target images have the same shape.",
                    f"The shapes mismatch. "
                    f"The target image has shape {target.shape}, while the interpretation target image has shape {interpretation_target.shape}."
                )
        """

        feature_interpretation_dict = {}

        for feature in template.features:

            feature_class = feature.get_feature_class()

            if feature_class is None:
                continue

            feature_img = supervision_result.warp_feature(feature, interpretation_target)
            feature_img_mutated = feature.apply_mutators_to_image(feature_img)
            interpretation = feature.interpret_image(feature_img_mutated)

            feature_interpretation_dict[feature.identifier] = interpretation

        return ExternalInterpretationResult(template, feature_interpretation_dict)



import strictyaml as yml

from officialeye._internal.context.singleton import get_internal_afi, get_internal_context

# noinspection PyProtectedMember
from officialeye._internal.feedback.verbosity import Verbosity
from officialeye._internal.template.internal_template import InternalTemplate
from officialeye._internal.template.schema.schema import generate_template_schema
from officialeye.error.errors.template import ErrTemplateInvalidSyntax

_oe_template_schema = generate_template_schema()


def _strict_yaml_error_to_syntax_error(error: yml.YAMLError, /, *, path: str) -> ErrTemplateInvalidSyntax:

    return ErrTemplateInvalidSyntax(
        f"while loading template configuration file at '{path}'.",
        "Could not parse the configuration file due to invalid syntax or encoding.",
        str(error).replace("<unicode string>", path)
    )


def _do_load_template(path: str, /) -> InternalTemplate:
    global _oe_template_schema

    with open(path, "r") as fh:
        raw_data = fh.read()

    try:
        yaml_document = yml.load(raw_data, schema=_oe_template_schema)
    except yml.YAMLError as err:
        raise _strict_yaml_error_to_syntax_error(err, path=path) from err

    data = yaml_document.data

    template = InternalTemplate(data, path)

    get_internal_afi().info(Verbosity.DEBUG, f"Loaded template: [b]{template}[/]")

    return template


def load_template(path: str, /) -> InternalTemplate:
    """
    Loads a template from a file located at the specified path.

    Arguments:
        path: The path to the YAML template configuration file.

    Returns:
        Loaded template.

    Raises:
        OEError: In case there has been an error validating the correctness of the template.
    """

    template = get_internal_context().get_template_by_path(path)

    if template is not None:
        get_internal_afi().info(Verbosity.DEBUG, f"Template at path '{path}' has already been loaded and cached, reusing it!")
        return template

    get_internal_afi().info(Verbosity.DEBUG, f"Template at path '{path}' has not yet been loaded, loading it.")

    return _do_load_template(path)



from __future__ import annotations

from concurrent.futures import Future
from typing import TYPE_CHECKING, Iterable

from officialeye._api.image import IImage
from officialeye._api.template.template_interface import ITemplate

# noinspection PyProtectedMember
from officialeye._internal.api.load import template_load

# noinspection PyProtectedMember
from officialeye._internal.template.external_template import ExternalTemplate

if TYPE_CHECKING:
    from officialeye._api.context import Context
    from officialeye._api.template.feature import IFeature
    from officialeye._api.template.keypoint import IKeypoint
    from officialeye._api.template.supervision_result import ISupervisionResult


class Template(ITemplate):

    def __init__(self, context: Context, /, *, path: str):
        super().__init__()

        self._context = context
        self._path = path

        # None indicates that the template has not yet been loaded
        self._external_template: ExternalTemplate | None = None

    def load(self) -> None:
        """
        Loads the template into memory for further processing.

        If you prefer lazy-evaluation, do not call this method.
        Instead, run the desired operations with the template, and the necessary resources will be loaded on-the-fly.
        Use this method only if you really want to preload the template.
        """

        if self._external_template is not None:
            # the template has already been loaded, nothing to do
            return

        # noinspection PyProtectedMember
        future = self._context._submit_task(template_load, "Loading template...", self._path)

        self._external_template = future.result()

        assert self._external_template is not None
        assert isinstance(self._external_template, ExternalTemplate)

    def detect_async(self, /, *, target: IImage) -> Future:
        self.load()
        return self._external_template.detect_async(target=target)

    def detect(self, /, **kwargs) -> ISupervisionResult:
        self.load()
        return self._external_template.detect(**kwargs)

    def get_image(self) -> IImage:
        self.load()
        return self._external_template.get_image()

    def get_mutated_image(self) -> IImage:
        self.load()
        return self._external_template.get_mutated_image()

    @property
    def identifier(self) -> str:
        self.load()
        return self._external_template.identifier

    @property
    def name(self) -> str:
        self.load()
        return self._external_template.name

    @property
    def width(self) -> int:
        self.load()
        return self._external_template.width

    @property
    def height(self) -> int:
        self.load()
        return self._external_template.height

    @property
    def keypoints(self) -> Iterable[IKeypoint]:
        self.load()
        return self._external_template.keypoints

    @property
    def features(self) -> Iterable[IFeature]:
        self.load()
        return self._external_template.features

    def get_feature(self, feature_id: str, /) -> IFeature | None:
        self.load()
        return self._external_template.get_feature(feature_id)

    def get_keypoint(self, keypoint_id: str, /) -> IKeypoint | None:
        self.load()
        return self._external_template.get_keypoint(keypoint_id)



from __future__ import annotations

from concurrent.futures import ALL_COMPLETED
from typing import TYPE_CHECKING, List

from officialeye._api.future import Future, wait
from officialeye._api.template.supervision_result import ISupervisionResult

# noinspection PyProtectedMember
from officialeye._internal.feedback.verbosity import Verbosity

# noinspection PyProtectedMember
from officialeye._internal.template.external_supervision_result import ExternalSupervisionResult
from officialeye.error.error import OEError
from officialeye.error.errors.internal import ErrInternal
from officialeye.error.errors.supervision import ErrSupervisionCorrespondenceNotFound

if TYPE_CHECKING:
    from officialeye._api.context import Context
    from officialeye._api.image import IImage
    from officialeye._api.template.template_interface import ITemplate


def detect(context: Context, *templates: ITemplate, target: IImage) -> ISupervisionResult:

    futures: List[Future] = [
        template.detect_async(target=target) for template in templates
    ]

    done, not_done = wait(futures, return_when=ALL_COMPLETED)

    if len(not_done) > 0:
        # noinspection PyProtectedMember
        context._get_afi().warn(Verbosity.DEBUG, "Some template analysis futures were not completed.")

    regular_errors: List[OEError] = []

    best_result: ISupervisionResult | None = None
    best_result_score: float = -1.0

    for completed_future in done:

        assert isinstance(completed_future, Future)

        if completed_future.cancelled():
            # noinspection PyProtectedMember
            context._get_afi().warn(Verbosity.DEBUG, "A template analysis future was cancelled.")
            continue

        error = completed_future.exception()

        if error is not None:
            # there has been an error during the execution of the future
            # the cause of the error might have been critical, in which case we should raise an exception immediately,
            # or it might be, for example, due to one of the templates not matching the image at all, which is regular behavior
            # therefore, we need to distinguish between a critical error and a regular one

            if not isinstance(error, OEError):
                err = ErrInternal(
                    "while analyzing target image against multiple templates.",
                    "One of the individual analysis workers has crashed due to an external error."
                )
                err.add_external_cause(error)
                raise err

            assert isinstance(error, OEError)

            if error.is_regular:
                # noinspection PyProtectedMember
                context._get_afi().warn(
                    Verbosity.DEBUG,
                    f"A template analysis worker has returned a regular error {error.code} ({error.code_text})."
                )
                regular_errors.append(error)
                continue

            # we are dealing with a non-regular OfficialEye error
            raise error

        result = completed_future.result()
        assert result is not None
        assert isinstance(result, ExternalSupervisionResult)

        # noinspection PyProtectedMember
        context._get_afi().info(Verbosity.DEBUG, f"Template analysis worker yielded a result with score {result.score}.")

        if result.score > best_result_score:
            best_result_score = result.score
            best_result = result

    if best_result is None:
        error = ErrSupervisionCorrespondenceNotFound(
            "while processing the target image analysis results.",
            "Could not establish correspondence of the image with any of the templates provided."
        )

        for worker_error in regular_errors:
            error.add_cause(worker_error)

        raise error

    return best_result



from __future__ import annotations

from typing import TYPE_CHECKING, Iterable, List

# noinspection PyProtectedMember
from officialeye._api.template.feature import IFeature
from officialeye._internal.api_implementation import IApiInterfaceImplementation
from officialeye._internal.template.region import ExternalRegion

if TYPE_CHECKING:
    # noinspection PyProtectedMember
    from officialeye._api.context import Context

    # noinspection PyProtectedMember
    from officialeye._api.mutator import IMutator
    from officialeye._internal.template.external_template import ExternalTemplate
    from officialeye._internal.template.internal_feature import InternalFeature


class ExternalFeature(ExternalRegion, IFeature, IApiInterfaceImplementation):

    def __init__(self, internal_feature: InternalFeature, external_template: ExternalTemplate, /):
        super().__init__(internal_feature, external_template)

        self._mutators: List[IMutator] = list(internal_feature.get_mutators())

    def get_mutators(self) -> Iterable[IMutator]:
        return self._mutators

    def set_api_context(self, context: Context, /) -> None:
        # no methods of this class require any contextual information to work, nothing to do
        pass

    def clear_api_context(self) -> None:
        # no methods of this class require any contextual information to work, nothing to do
        pass



from __future__ import annotations

from typing import TYPE_CHECKING

from officialeye._internal.api_implementation import IApiInterfaceImplementation
from officialeye._internal.template.external_template import ExternalTemplate
from officialeye._internal.template.shared_matching_result import SharedMatchingResult

if TYPE_CHECKING:
    # noinspection PyProtectedMember
    from officialeye._api.context import Context
    from officialeye._internal.template.internal_matching_result import InternalMatchingResult


class ExternalMatchingResult(SharedMatchingResult, IApiInterfaceImplementation):
    """
    Representation of the matching result, designed to be used by the main process.
    For this reason, it is essential that this class is picklable.
    """

    def __init__(self, internal_matching_result: InternalMatchingResult, external_template: ExternalTemplate, /):
        super().__init__(internal_matching_result.template)

        self._template = external_template

    @property
    def template(self) -> ExternalTemplate:
        return self._template

    def set_api_context(self, context: Context, /) -> None:
        self._template.set_api_context(context)

    def clear_api_context(self) -> None:
        self._template.clear_api_context()



"""
This module provides a set of functions connecting the API interface with the internal implementation interface.
In other words, the functions of this module form a low-level API that should be called internally in the actual public API.
"""


from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Iterable

import numpy as np

from officialeye._api.config import MatcherConfig
from officialeye._api.template.keypoint import IKeypoint
from officialeye._api.template.match import IMatch

if TYPE_CHECKING:
    from officialeye._api.template.template import ITemplate
    from officialeye.types import ConfigDict


class IMatcher(ABC):

    @property
    @abstractmethod
    def config(self) -> MatcherConfig:
        raise NotImplementedError()

    @abstractmethod
    def setup(self, target: np.ndarray, template: ITemplate, /) -> None:
        raise NotImplementedError()

    @abstractmethod
    def match(self, keypoint: IKeypoint, /) -> None:
        raise NotImplementedError()

    @abstractmethod
    def get_matches_for_keypoint(self, keypoint: IKeypoint, /) -> Iterable[IMatch]:
        raise NotImplementedError()


class Matcher(IMatcher, ABC):

    def __init__(self, matcher_id: str, config_dict: ConfigDict, /):
        super().__init__()

        self.matcher_id = matcher_id

        self._config = MatcherConfig(config_dict, matcher_id)

    @property
    def config(self) -> MatcherConfig:
        return self._config



from __future__ import annotations

from typing import TYPE_CHECKING

import cv2
import numpy as np

# noinspection PyProtectedMember
from officialeye._api.mutator import Mutator
from officialeye.error.errors.template import ErrTemplateInvalidMutator

if TYPE_CHECKING:
    from officialeye.types import ConfigDict


class RotateMutator(Mutator):
    """
    Mutator that rotates the image clockwise by the specified angle.
    """

    MUTATOR_ID = "rotate"

    def __init__(self, config: ConfigDict, /):
        super().__init__(RotateMutator.MUTATOR_ID, config)

        def _angle_preprocessor(angle_text: str) -> int:
            angle = int(angle_text)

            if angle not in (0, 90, 180, 270):
                raise ErrTemplateInvalidMutator(
                    f"while loading mutator '{self.mutator_id}'.",
                    f"The 'angle' parameter must be a multiple of 90 degrees, got {angle}."
                )

            return angle

        self._angle = self.config.get("angle", value_preprocessor=_angle_preprocessor)

    def mutate(self, img: np.ndarray, /) -> np.ndarray:

        if self._angle == 0:
            # we do not need to rotate the image at all
            return img

        if self._angle == 90:
            cv2_rotate_code = cv2.ROTATE_90_CLOCKWISE
        elif self._angle == 180:
            cv2_rotate_code = cv2.ROTATE_180
        elif self._angle == 270:
            cv2_rotate_code = cv2.ROTATE_90_COUNTERCLOCKWISE
        else:
            raise AssertionError()

        return cv2.rotate(img, cv2_rotate_code)



from abc import ABC, abstractmethod

from officialeye._api.template.region import IRegion


class IKeypoint(IRegion, ABC):

    @property
    @abstractmethod
    def matches_min(self) -> int:
        raise NotImplementedError()

    @property
    @abstractmethod
    def matches_max(self) -> int:
        raise NotImplementedError()

    def __str__(self) -> str:
        return f"Keypoint '{self.identifier}'"



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

# noinspection PyProtectedMember
from officialeye._api.mutator import Mutator

if TYPE_CHECKING:
    from officialeye.types import ConfigDict


class CropMutator(Mutator):

    MUTATOR_ID = "crop"

    def __init__(self, config_dict: ConfigDict, /):
        super().__init__(CropMutator.MUTATOR_ID, config_dict)

        self._x = self.config.get("x", default=0, value_preprocessor=int)
        self._y = self.config.get("y", default=0, value_preprocessor=int)
        self._w = self.config.get("w", value_preprocessor=int)
        self._h = self.config.get("h", value_preprocessor=int)

    def mutate(self, img: np.ndarray, /) -> np.ndarray:
        return img[self._y:self._y + self._h, self._x:self._x + self._w]



"""
Root module.
"""

# disable unused imports ruff check
# ruff: noqa: F401

# Config
# noinspection PyProtectedMember
from officialeye._api.config import Config, InterpretationConfig, MatcherConfig, MutatorConfig, SupervisorConfig

# Context
# noinspection PyProtectedMember
from officialeye._api.context import Context

# Misc
# noinspection PyProtectedMember
from officialeye._api.future import Future, wait

# Image-processing
# noinspection PyProtectedMember
from officialeye._api.image import IImage, Image

# Mutators
# noinspection PyProtectedMember
from officialeye._api.mutator import IMutator, Mutator

# noinspection PyProtectedMember
from officialeye._api.template.feature import IFeature

# Interpretation-related imports
# noinspection PyProtectedMember
from officialeye._api.template.interpretation import IInterpretation, Interpretation

# noinspection PyProtectedMember
from officialeye._api.template.interpretation_result import IInterpretationResult

# noinspection PyProtectedMember
from officialeye._api.template.keypoint import IKeypoint

# Matching-related imports
# noinspection PyProtectedMember
from officialeye._api.template.match import IMatch, Match

# noinspection PyProtectedMember
from officialeye._api.template.matcher import IMatcher, Matcher

# noinspection PyProtectedMember
from officialeye._api.template.matching_result import IMatchingResult

# Regions, features and keypoints
# noinspection PyProtectedMember
from officialeye._api.template.region import IRegion, Region

# noinspection PyProtectedMember
from officialeye._api.template.supervision_result import ISupervisionResult, SupervisionResult

# Supervision-related imports
# noinspection PyProtectedMember
from officialeye._api.template.supervisor import ISupervisor, Supervisor

# Template-related
# noinspection PyProtectedMember
from officialeye._api.template.template import ITemplate, Template



from abc import ABC, abstractmethod
from typing import Any, Iterable

from officialeye._api.future import Future
from officialeye._api.image import IImage
from officialeye._api.template.feature import IFeature
from officialeye._api.template.keypoint import IKeypoint
from officialeye._api.template.supervision_result import ISupervisionResult


class ITemplate(ABC):

    def __init__(self):
        super().__init__()

    @abstractmethod
    def load(self) -> None:
        """
        Loads the template into memory for further processing.

        If you prefer lazy-evaluation, do not call this method.
        Instead, run the desired operations with the template, and the necessary resources will be loaded on-the-fly.
        Use this method only if you really want to preload the template.
        """

        raise NotImplementedError()

    @abstractmethod
    def detect_async(self, /, *, target: IImage) -> Future:
        raise NotImplementedError()

    @abstractmethod
    def detect(self, /, **kwargs) -> ISupervisionResult:
        raise NotImplementedError()

    @abstractmethod
    def get_image(self) -> IImage:
        raise NotImplementedError()

    @abstractmethod
    def get_mutated_image(self) -> IImage:
        raise NotImplementedError()

    @property
    @abstractmethod
    def identifier(self) -> str:
        raise NotImplementedError()

    @property
    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError()

    @property
    @abstractmethod
    def width(self) -> int:
        raise NotImplementedError()

    @property
    @abstractmethod
    def height(self) -> int:
        raise NotImplementedError()

    @property
    @abstractmethod
    def keypoints(self) -> Iterable[IKeypoint]:
        raise NotImplementedError()

    @property
    @abstractmethod
    def features(self) -> Iterable[IFeature]:
        raise NotImplementedError()

    @abstractmethod
    def get_feature(self, feature_id: str, /) -> IFeature | None:
        raise NotImplementedError()

    @abstractmethod
    def get_keypoint(self, keypoint_id: str, /) -> IKeypoint | None:
        raise NotImplementedError()

    def __str__(self) -> str:
        return f"Template '{self.identifier}'."

    def __eq__(self, o: Any) -> bool:

        if not isinstance(o, ITemplate):
            return False

        return self.identifier == o.identifier

    def __hash__(self):
        return hash(self.identifier)



from __future__ import annotations

from concurrent.futures import Future
from contextlib import contextmanager
from multiprocessing import Pipe

# noinspection PyProtectedMember
from multiprocessing.connection import Connection, wait
from threading import Lock, Thread
from types import TracebackType
from typing import Any, Dict, List, Set, Tuple

from rich.console import Console, ConsoleRenderable
from rich.panel import Panel
from rich.progress import Progress, SpinnerColumn, TaskID, TextColumn
from rich.theme import Theme
from rich.traceback import Traceback

# noinspection PyProtectedMember
from officialeye._internal.context.feedback import InternalFeedbackInterface, IPCMessageType

# noinspection PyProtectedMember
from officialeye._internal.feedback.abstract import AbstractFeedbackInterface

# noinspection PyProtectedMember
from officialeye._internal.feedback.verbosity import Verbosity
from officialeye.error.error import OEError
from officialeye.error.errors.general import ErrOperationNotSupported
from officialeye.error.errors.internal import ErrInternal

_THEME_TAG_INFO = "info"
_THEME_TAG_INFO_VERBOSE = "infov"
_THEME_TAG_DEBUG = "debug"
_THEME_TAG_DEBUG_VERBOSE = "debugv"

_THEME_TAG_WARN = "warn"
_THEME_TAG_ERR = "err"

_THEME: Dict[str, str] = {
    _THEME_TAG_INFO: "bold green",
    _THEME_TAG_INFO_VERBOSE: "bold cyan",
    _THEME_TAG_DEBUG: "bold blue",
    _THEME_TAG_DEBUG_VERBOSE: "bold purple",
    _THEME_TAG_WARN: "bold yellow",
    _THEME_TAG_ERR: "bold red"
}

_THEME_MAP: Dict[Verbosity, str] = {
    Verbosity.INFO: _THEME_TAG_INFO,
    Verbosity.INFO_VERBOSE: _THEME_TAG_INFO_VERBOSE,
    Verbosity.DEBUG: _THEME_TAG_DEBUG,
    Verbosity.DEBUG_VERBOSE: _THEME_TAG_DEBUG_VERBOSE
}


def _wrap_exception(exception_value: BaseException, /) -> OEError:

    if isinstance(exception_value, OEError):
        return exception_value

    if isinstance(exception_value, BaseException):
        oe_error = ErrInternal(
            "while leaving the CLI context.",
            "An internal error occurred.",
        )
        oe_error.add_external_cause(exception_value)
        return oe_error

    return ErrInternal(
        "while determining the raised exception type.",
        "Could not decide how to handle the raised error."
    )


def _child_listener(listener: _ChildrenListener, /):

    # a known child is a child that the listener thread is either already activaly listening to, or has already stopped listening to due to
    # the receival of a message over IPC, indicating that the child is done with the work
    known_children: Set[int] = set()

    while True:

        # first, we wait for one of the children to send some message
        with listener.children_lock:
            if len(listener.children) == 0:
                break

            # determine whether there are children the children listener thread does not know about
            # in this case, those children should be found and added to the known list,
            # and the corresponding lock should be acquired, indicating that the connection with the current
            # child has not yet been shut down gracefully
            for child_id in listener.children:

                if child_id in known_children:
                    continue

                child = listener.children[child_id]

                # indicate that we are listening to that child, and that once it is done with its work,
                # our handling of messages that might still be pending on the IPC, is to be respected
                child.is_being_listened_to.acquire()

                # now we know the child
                known_children.add(child_id)

            connections = [
                listener.children[child_id].connection for child_id in listener.children
            ]

            assert len(connections) >= 1

        # TODO: consider introducing a mechanism allowing one to let the child listener thread know about the change in the children dictionary
        # TODO: this will improve performance, because there will be no need to wait for the timeout to expire
        # TODO: this idea can be implemented, for example, by introducing a new dummy connection designed only to communicate 'refresh' messages
        # TODO: with the child listener thread
        wait(connections, timeout=1.0)

        messages_to_handle: List[Tuple[Any, _Child]] = []

        with listener.children_lock:
            if len(listener.children) == 0:
                break

            for child_id in listener.children:
                child = listener.children[child_id]

                if not child.connection.poll():
                    continue

                message = child.connection.recv()

                messages_to_handle.append((message, child))
                    
        for message, child in messages_to_handle:
            listener.handle_message(message, child)


class _Child:

    def __init__(self, child_id: int, task_id: TaskID, connection: Connection, /):
        self.child_id = child_id
        self.task_id = task_id
        self.connection = connection
        self.is_being_listened_to = Lock()


# noinspection PyProtectedMember
class _ChildrenListener:

    def __init__(self, terminal_ui, /):

        self._terminal_ui = terminal_ui

        self._progress = Progress(
            SpinnerColumn(),
            TextColumn("{task.description}"),
            TextColumn("[cyan]{task.fields[status]}"),
            console=self._terminal_ui._console,
            auto_refresh=True,
            disable=self._terminal_ui._verbosity == Verbosity.QUIET,
            transient=True
        )

        self.children: Dict[int, _Child] = {}
        self.children_lock = Lock()

        self._children_listener: Thread | None = None

    def handle_message(self, message: tuple, child: _Child, /):

        message_type, args, kwargs = message

        is_task_done: bool = False

        with self._terminal_ui.as_author(child.child_id):
            if message_type == IPCMessageType.ECHO:
                self._terminal_ui.echo(*args, **kwargs)
            elif message_type == IPCMessageType.INFO:
                self._terminal_ui.info(*args, **kwargs)
            elif message_type == IPCMessageType.WARN:
                self._terminal_ui.warn(*args, **kwargs)
            elif message_type == IPCMessageType.ERROR:
                self._terminal_ui.error(*args, **kwargs)
            elif message_type == IPCMessageType.UPDATE_STATUS:
                new_status_text: str = args[0]
                assert isinstance(new_status_text, str)
                self._progress.update(child.task_id, status=new_status_text)
            elif message_type == IPCMessageType.TASK_DONE:
                is_task_done = True
            else:
                raise AssertionError("Unknown IPC message type received by parent process.")

        if is_task_done:

            task_done_successfully: bool = args[0]
            assert isinstance(task_done_successfully, bool)

            self._terminal_ui.info(
                Verbosity.DEBUG,
                f"Child has indicated that the task is done (success={task_done_successfully}), "
                "releasing the corresponding lock to enable graceful shutdown of the child listener."
            )

            if task_done_successfully:
                self._progress.update(child.task_id, completed=100, status="[green]:heavy_check_mark: Success![/]")
            else:
                self._progress.update(child.task_id, completed=100, status="[red]:heavy_multiplication_x: Error![/]")

            child.is_being_listened_to.release()

    def listen_to(self, child_id: int, connection: Connection, description: str, /):

        # create a new task associated with the child
        task_id = self._progress.add_task(description, status="")

        with self.children_lock:
            assert child_id not in self.children, "Child ID is not unique."
            self.children[child_id] = _Child(child_id, task_id, connection)

        if self._children_listener is None:
            # we have added the first child. therefore, the progress bar needs to be started.
            self._progress.start()

            # we need to also start a thread listening for messages from children
            self._children_listener = Thread(target=_child_listener, name="Child Process Listener", args=(self,))
            self._children_listener.start()

    def stop_listening_to(self, child_id: int, /):

        # we first attempt to wait until the child listener thread has done processing all messages sent by the child
        # in other words we want to try and gracefully stop listening to the child

        with self.children_lock:

            if child_id not in self.children:
                self._terminal_ui.warn(
                    Verbosity.DEBUG,
                    f"Could not stop listening for child {child_id} because it could not be found among children the main process is listening to."
                )
                return

            _child = self.children[child_id]

            self._terminal_ui.info(
                Verbosity.DEBUG,
                f"Waiting for all messages from child {child_id} to be processed by the child listener thread."
            )

            child_lock = _child.is_being_listened_to

        if child_lock.acquire(blocking=True, timeout=4):
            child_lock.release()

            self._terminal_ui.info(
                Verbosity.DEBUG,
                f"The child listener thread indicated that it has processed all messages from child {child_id}."
            )
        else:
            self._terminal_ui.warn(
                Verbosity.DEBUG,
                f"The child listener thread did not indicate that it has processed all messages from child {child_id}!"
            )

        # we now proceed with removing the child completely

        last_child_removed = False

        with self.children_lock:

            if child_id not in self.children:
                self._terminal_ui.warn(
                    Verbosity.DEBUG,
                    f"Could not stop listening for child {child_id} because it could not be found among children the main process is listening to."
                )
                return

            child = self.children[child_id]

            child.connection.close()

            del self.children[child_id]

            if len(self.children) == 0:
                # we have removed the last child
                last_child_removed = True

        if last_child_removed:
            self._terminal_ui.info(Verbosity.DEBUG_VERBOSE, "Last child removed, stopping the child listener and the progress bar.")

            # stop the thread listening for messages from children
            if self._children_listener is not None:
                self._terminal_ui.info(Verbosity.DEBUG_VERBOSE, "Joining the children listener thread.")

                self._children_listener.join()
                self._children_listener = None

                self._terminal_ui.info(Verbosity.DEBUG_VERBOSE, "Children listener thread successfully joined.")

            # stop the progress bar
            self._terminal_ui.info(Verbosity.DEBUG_VERBOSE, "Stopping the progress bar due to removal of last child.")
            self._progress.stop()
            self._terminal_ui.info(Verbosity.DEBUG_VERBOSE, "Stopped the progress bar due to removal of last child.")

    def remove_all_children(self):

        while True:

            with self.children_lock:
                children_to_be_removed = list(self.children.keys())

            if len(children_to_be_removed) == 0:
                break

            for child_id in children_to_be_removed:
                self.stop_listening_to(child_id)

    def dispose(self):
        self._terminal_ui.info(Verbosity.DEBUG_VERBOSE, "Dispoing child listener...")
        self.remove_all_children()


class TerminalUI(AbstractFeedbackInterface):

    def __init__(self, verbosity: Verbosity, /):
        super().__init__(verbosity)

        self._console = Console(highlight=False, theme=Theme(_THEME))
        self._err_console = Console(stderr=True, theme=Theme(_THEME))

        self._children_listener: _ChildrenListener = _ChildrenListener(self)
        self._fork_counter: int = 0

        self._last_printed_message_author: int | None = None

    def get_console(self) -> Console:
        return self._console

    def _print_message_authors(self) -> bool:
        return self._verbosity >= Verbosity.DEBUG

    @contextmanager
    def as_author(self, author: int):

        print_authors = self._print_message_authors()

        if print_authors and self._last_printed_message_author != author:
            # the same author prints the message
            self._console.rule(f"Messages by worker #{author}")

        self._last_printed_message_author = None

        yield self

        if print_authors:
            self._last_printed_message_author = author

    def echo(self, verbosity: Verbosity, message: str | ConsoleRenderable = "", /, *, err: bool = False, **kwargs: Any) -> None:

        assert verbosity != Verbosity.QUIET

        if self._last_printed_message_author is not None:
            self._console.rule("Messages by the main process")
            self._last_printed_message_author = None

        if self._verbosity < verbosity:
            return

        console = self._err_console if err else self._console

        if not console.is_interactive:
            kwargs.setdefault("crop", False)
            kwargs.setdefault("overflow", "ignore")

        console.print(message, **kwargs)

    def info(self, verbosity: Verbosity, message: str, /) -> None:
        assert verbosity != Verbosity.QUIET
        _tag = _THEME_MAP[verbosity]
        self.echo(verbosity, f"[{_tag}]INFO [/] {message}", highlight=True)

    def warn(self, verbosity: Verbosity, message: str, /) -> None:
        assert verbosity != Verbosity.QUIET
        self.echo(verbosity, f"[{_THEME_TAG_WARN}]WARN [/] {message}")

    def error(self, verbosity: Verbosity, message: str, /) -> None:
        assert verbosity != Verbosity.QUIET
        self.echo(verbosity, f"[{_THEME_TAG_ERR}]ERROR[/] {message}", err=True)

    def update_status(self, new_status_text: str, /) -> None:
        raise ErrOperationNotSupported(
            "while updating status of a task.",
            "The terminal UI does not support this operation."
        )

    def _print_oe_error(self, error: OEError, /, *, verbosity: Verbosity = Verbosity.INFO):

        self.error(verbosity, f"Error {error.code} ('{error.code_text}') occurred in module '{error.module}' {error.while_text}")
        self.error(verbosity, f"Problem: {error.problem_text}")

        error_details = error.get_details()

        if error_details is not None and verbosity != Verbosity.QUIET:
            detail_panel = Panel(error_details, title="Error details", expand=False, border_style="red", highlight=True)
            self._err_console.print(detail_panel)

        causes = error.get_causes()
        external_causes = error.get_external_causes()

        if len(causes) + len(external_causes) > 0:
            self._err_console.rule("The above error has been caused by the errors listed below")

        for cause in causes:
            self._print_oe_error(cause, verbosity=verbosity)

        for external_cause in external_causes:
            rich_traceback = Traceback.from_exception(external_cause.__class__, external_cause, None)
            self._err_console.print(rich_traceback)

    def handle_uncaught_error(self, exception_type: any, exception_value: BaseException, traceback: TracebackType, /):

        # remove all children to make sure that the progress bar dissapears and
        # does not interfere with the printing of the error to the terminal
        self._children_listener.remove_all_children()

        self._print_oe_error(_wrap_exception(exception_value))

        if self._verbosity >= Verbosity.DEBUG_VERBOSE:
            # self._err_console.rule("Error details")
            rich_traceback = Traceback.from_exception(exception_type, exception_value, traceback)
            self._err_console.print(rich_traceback)

    def dispose(self, exception_type: any = None, exception_value: BaseException | None = None, traceback: TracebackType | None = None) -> None:
        self._children_listener.dispose()

    def fork(self, description: str, /) -> AbstractFeedbackInterface:

        self.info(Verbosity.DEBUG_VERBOSE, "AbstractFeedbackInterface: fork()")

        rx, tx = Pipe(duplex=False)

        assert isinstance(rx, Connection)
        assert isinstance(tx, Connection)

        self._fork_counter += 1
        child_id = self._fork_counter

        child = InternalFeedbackInterface(self._verbosity, child_id, tx)

        self._children_listener.listen_to(child_id, rx, description)

        return child

    def join(self, child: AbstractFeedbackInterface, future: Future, /) -> None:

        assert isinstance(child, InternalFeedbackInterface), "Invalid child type"

        child_id = child.get_child_id()

        self.info(Verbosity.DEBUG_VERBOSE, f"AbstractFeedbackInterface: join() of child #{child_id}")

        self._children_listener.stop_listening_to(child_id)



from __future__ import annotations

import os
import random
from typing import TYPE_CHECKING, Dict, Iterable, List

import numpy as np

# noinspection PyProtectedMember
from officialeye._api.future import Future

# noinspection PyProtectedMember
from officialeye._api.image import IImage

# noinspection PyProtectedMember
from officialeye._api.template.match import IMatch

# noinspection PyProtectedMember
from officialeye._api.template.supervisor import ISupervisor

# noinspection PyProtectedMember
from officialeye._api.template.template import ITemplate
from officialeye._internal.context.singleton import get_internal_afi, get_internal_context

# noinspection PyProtectedMember
from officialeye._internal.feedback.verbosity import Verbosity
from officialeye._internal.template.feature_class.loader import load_template_feature_classes
from officialeye._internal.template.feature_class.manager import FeatureClassManager
from officialeye._internal.template.image import InternalImage
from officialeye._internal.template.internal_feature import InternalFeature
from officialeye._internal.template.internal_matching_result import InternalMatchingResult
from officialeye._internal.template.internal_supervision_result import InternalSupervisionResult
from officialeye._internal.template.keypoint import InternalKeypoint
from officialeye._internal.template.utils import load_mutator_from_dict
from officialeye._internal.timer import Timer
from officialeye.error.errors.general import ErrInvalidIdentifier, ErrOperationNotSupported
from officialeye.error.errors.supervision import ErrSupervisionCorrespondenceNotFound
from officialeye.error.errors.template import ErrTemplateInvalidFeature, ErrTemplateInvalidKeypoint

if TYPE_CHECKING:
    # noinspection PyProtectedMember
    from officialeye._api.mutator import IMutator

    # noinspection PyProtectedMember
    from officialeye._api.template.matcher import IMatcher

    # noinspection PyProtectedMember
    from officialeye._api.template.supervision_result import ISupervisionResult
    from officialeye.types import ConfigDict


# TODO: refactor this into an enum
_SUPERVISION_RESULT_FIRST = "first"
_SUPERVISION_RESULT_RANDOM = "random"
_SUPERVISION_RESULT_BEST_MSE = "best_mse"
_SUPERVISION_RESULT_BEST_SCORE = "best_score"


class InternalTemplate(ITemplate):

    def __init__(self, yaml_dict: Dict[str, any], path_to_template: str, /):
        super().__init__()

        self._path_to_template = path_to_template

        self._template_id = yaml_dict["id"]
        self._name = yaml_dict["name"]
        self._source = yaml_dict["source"]

        self._height, self._width, _ = self.get_image().load().shape

        self._source_mutators: List[IMutator] = [
            load_mutator_from_dict(mutator_dict) for mutator_dict in yaml_dict["mutators"]["source"]
        ]

        self._target_mutators: List[IMutator] = [
            load_mutator_from_dict(mutator_dict) for mutator_dict in yaml_dict["mutators"]["target"]
        ]

        self._keypoints: Dict[str, InternalKeypoint] = {}
        self._features: Dict[str, InternalFeature] = {}

        for keypoint_id in yaml_dict["keypoints"]:
            keypoint_dict = yaml_dict["keypoints"][keypoint_id]
            keypoint_dict["id"] = keypoint_id
            keypoint = InternalKeypoint(self.identifier, keypoint_dict)

            if keypoint.identifier in self._keypoints:
                raise ErrTemplateInvalidKeypoint(
                    f"while initializing keypoint '{keypoint_id}' of template '{self.identifier}'",
                    f"There is already a keypoint with the same identifier '{keypoint.identifier}'."
                )

            if keypoint.identifier in self._features:
                raise ErrTemplateInvalidKeypoint(
                    f"while initializing keypoint '{keypoint_id}' of template '{self.identifier}'",
                    f"There is already a feature with the same identifier '{keypoint.identifier}'."
                )

            self._keypoints[keypoint.identifier] = keypoint

        self._matching = yaml_dict["matching"]
        self._supervision = yaml_dict["supervision"]

        # load feature classes
        self._feature_class_manager = load_template_feature_classes(yaml_dict["feature_classes"], self.identifier)

        # load features
        for feature_id in yaml_dict["features"]:
            feature_dict = yaml_dict["features"][feature_id]
            feature_dict["id"] = feature_id
            feature = InternalFeature(self.identifier, feature_dict)

            if feature.identifier in self._keypoints:
                raise ErrTemplateInvalidFeature(
                    f"while initializing feature '{feature_id}' of template '{self.identifier}'",
                    f"There is already a keypoint with the same identifier '{feature.identifier}'."
                )

            if feature.identifier in self._features:
                raise ErrTemplateInvalidFeature(
                    f"while initializing feature '{feature_id}' of template '{self.identifier}'",
                    f"There is already a feature with the same identifier '{feature.identifier}'."
                )

            self._features[feature.identifier] = feature

        get_internal_context().add_template(self)

    def get_source_mutators(self) -> Iterable[IMutator]:
        return self._source_mutators

    def get_target_mutators(self) -> Iterable[IMutator]:
        return self._target_mutators

    def load(self) -> None:
        raise ErrOperationNotSupported(
            "while accessing an internal template instance.",
            "The way in which it was accessed is not supported."
        )

    def detect_async(self, /, *, target: IImage) -> Future:
        raise ErrOperationNotSupported(
            "while accessing an internal template instance.",
            "The way in which it was accessed is not supported."
        )

    def detect(self, /, **kwargs) -> ISupervisionResult:
        raise ErrOperationNotSupported(
            "while accessing an internal template instance.",
            "The way in which it was accessed is not supported."
        )

    def get_source_image_path(self) -> str:
        if os.path.isabs(self._source):
            return self._source
        path_to_template_dir = os.path.dirname(self._path_to_template)
        path = os.path.join(path_to_template_dir, self._source)
        return os.path.normpath(path)

    def get_image(self) -> IImage:
        return InternalImage(path=self.get_source_image_path())

    def get_mutated_image(self) -> IImage:
        img = self.get_image()
        img.apply_mutators(*self._source_mutators)
        return img

    @property
    def identifier(self) -> str:
        return self._template_id

    @property
    def name(self) -> str:
        return self._name

    @property
    def width(self) -> int:
        return self._width

    @property
    def height(self) -> int:
        return self._height

    @property
    def keypoints(self) -> Iterable[InternalKeypoint]:
        for keypoint_id in self._keypoints:
            yield self._keypoints[keypoint_id]

    @property
    def features(self) -> Iterable[InternalFeature]:
        for feature_id in self._features:
            yield self._features[feature_id]

    def validate(self):
        for feature in self.features:
            feature.validate_feature_class()

    def get_matcher(self, /) -> IMatcher:
        matcher_id = self._matching["engine"]

        matcher_configs: dict = self._matching["config"]
        matcher_config: ConfigDict = matcher_configs.get(matcher_id, {})

        return get_internal_context().get_matcher(matcher_id, matcher_config)

    def get_supervisor(self, /) -> ISupervisor:
        supervisor_id = self._supervision["engine"]
        supervisor_config_generic = self._supervision["config"]

        if supervisor_id in supervisor_config_generic:
            supervisor_config: ConfigDict = supervisor_config_generic[supervisor_id]
        else:
            get_internal_afi().warn(Verbosity.INFO, f"Could not find any configuration entries for the '{supervisor_id}' supervisor.")
            supervisor_config: ConfigDict = {}

        return get_internal_context().get_supervisor(supervisor_id, supervisor_config)

    def get_supervision_config(self) -> dict:
        return self._supervision["config"]

    def get_feature_classes(self) -> FeatureClassManager:
        return self._feature_class_manager

    def get_feature(self, feature_id: str, /) -> InternalFeature | None:

        if feature_id not in self._features:
            return None

        return self._features[feature_id]

    def get_keypoint(self, keypoint_id: str, /) -> InternalKeypoint | None:

        if keypoint_id not in self._keypoints:
            return None

        return self._keypoints[keypoint_id]

    def get_path(self) -> str:
        return self._path_to_template

    def _run_supervisor(self, keypoint_matching_result: InternalMatchingResult, /) -> InternalSupervisionResult | None:

        supervisor = self.get_supervisor()
        supervisor.setup(self, keypoint_matching_result)

        supervision_result_choice_engine = self._supervision["result"]
        results: List[InternalSupervisionResult] = [
            InternalSupervisionResult(supervision_result, self, keypoint_matching_result)
            for supervision_result in supervisor.supervise(self, keypoint_matching_result)
        ]

        if len(results) == 0:
            return None

        for result in results:
            get_internal_afi().info(
                Verbosity.INFO_VERBOSE,
                f"Got result with score {result.score} and error {result.get_weighted_mse()} from supervisor '{supervisor}'."
            )

        if supervision_result_choice_engine == _SUPERVISION_RESULT_FIRST:
            return results[0]

        if supervision_result_choice_engine == _SUPERVISION_RESULT_RANDOM:
            return random.choice(results)

        if supervision_result_choice_engine == _SUPERVISION_RESULT_BEST_MSE:

            best_result = results[0]
            best_result_mse = best_result.get_weighted_mse()

            for result_id, result in enumerate(results):
                result_mse = result.get_weighted_mse()

                get_internal_afi().info(Verbosity.INFO_VERBOSE, f"Result #{result_id + 1} has MSE {result_mse}.")

                if result_mse < best_result_mse:
                    best_result_mse = result_mse
                    best_result = result

            get_internal_afi().info(Verbosity.INFO_VERBOSE, f"Best result has MSE {best_result_mse}.")

            return best_result

        if supervision_result_choice_engine == _SUPERVISION_RESULT_BEST_SCORE:

            best_result = results[0]
            best_result_mse = best_result.get_weighted_mse()
            best_result_score = best_result.score

            for result_id, result in enumerate(results):
                result_score = result.score
                result_mse = result.get_weighted_mse()

                get_internal_afi().info(Verbosity.INFO_VERBOSE, f"Result #{result_id + 1} has score {result_score}.")

                if result_score > best_result_score or result_score == best_result_score and result_mse < best_result_mse:
                    best_result_mse = result_mse
                    best_result_score = result_score
                    best_result = result

            get_internal_afi().info(Verbosity.INFO_VERBOSE, f"Best result has score {best_result_score} and MSE {best_result_mse}.")

            return best_result

        raise ErrInvalidIdentifier(
            "while running supervisor.",
            f"Invalid supervision result choice engine '{supervision_result_choice_engine}'."
        )

    def do_detect(self, target: np.ndarray, /) -> InternalSupervisionResult:
        # find all patterns in the target image

        # prepare target image
        get_internal_afi().update_status("Preparing target image...")

        # apply mutators to the target image
        for mutator in self._target_mutators:
            target = mutator.mutate(target)

        get_internal_afi().update_status("Running matching phase...")

        _timer = Timer()

        with _timer:
            # start matching
            matcher: IMatcher = self.get_matcher()
            matcher.setup(target, self)

            for keypoint in self.keypoints:
                get_internal_afi().info(Verbosity.DEBUG, f"Running matcher '{matcher}' for keypoint '{keypoint.identifier}'.")
                assert isinstance(keypoint, InternalKeypoint)
                matcher.match(keypoint)

            keypoint_matching_result = InternalMatchingResult(self)

            for keypoint in self.keypoints:
                for match in matcher.get_matches_for_keypoint(keypoint):
                    assert isinstance(match, IMatch)
                    keypoint_matching_result.add_match(match)

            keypoint_matching_result.validate()
            assert keypoint_matching_result.get_total_match_count() > 0

        get_internal_afi().info(
            Verbosity.INFO,
            f"Matching succeeded in {_timer.get_real_time():.2f} seconds of real time "
            f"and {_timer.get_cpu_time():.2f} seconds of CPU time."
        )

        get_internal_afi().update_status("Running supervision phase...")

        with _timer:
            # run supervision to obtain correspondence between template and target regions
            supervision_result = self._run_supervisor(keypoint_matching_result)

        get_internal_afi().info(
            Verbosity.INFO,
            f"Supervision succeeded in {_timer.get_real_time():.2f} seconds of real time "
            f"and {_timer.get_cpu_time():.2f} seconds of CPU time."
        )

        if supervision_result is None:
            raise ErrSupervisionCorrespondenceNotFound(
                "while processing a supervision result.",
                f"Could not establish correspondence of the image with the '{self.identifier}' template."
            )

        get_internal_afi().info(
            Verbosity.DEBUG_VERBOSE,
            f"Supervision result: Delta = {supervision_result.delta} Delta' = {supervision_result.delta_prime} "
            f"M = {supervision_result.transformation_matrix}"
        )

        # TODO: visualizations
        """
        if get_internal_context().visualization_generation_enabled():
            supervision_result_visualizer = SupervisionResultVisualizer(supervision_result, target)
            visualization = supervision_result_visualizer.render()
            get_internal_context().export_image(visualization, file_name="matches.png")
        """

        return supervision_result

    def __str__(self):
        return f"{self.name} ({self._source}, {len(self._keypoints)} keypoints, {len(self._features)} features)"



from __future__ import annotations

import random
from typing import TYPE_CHECKING, Dict, Iterable, List

import numpy as np
import z3

# noinspection PyProtectedMember
from officialeye._api.template.match import IMatch

# noinspection PyProtectedMember
from officialeye._api.template.matching_result import IMatchingResult

# noinspection PyProtectedMember
from officialeye._api.template.supervision_result import SupervisionResult

# noinspection PyProtectedMember
from officialeye._api.template.supervisor import Supervisor

# noinspection PyProtectedMember
from officialeye._api.template.template_interface import ITemplate

# noinspection PyProtectedMember
from officialeye._internal.context.singleton import get_internal_afi

# noinspection PyProtectedMember
from officialeye._internal.feedback.verbosity import Verbosity
from officialeye.error.errors.supervision import ErrSupervisionInvalidEngineConfig

if TYPE_CHECKING:
    from officialeye.types import ConfigDict


class CombinatorialSupervisor(Supervisor):

    SUPERVISOR_ID = "combinatorial"

    def __init__(self, config_dict: ConfigDict, /):
        super().__init__(CombinatorialSupervisor.SUPERVISOR_ID, config_dict)

        # setup configuration
        def _min_match_factor_preprocessor(v: str) -> float:

            v = float(v)

            if v > 1.0:
                raise ErrSupervisionInvalidEngineConfig(
                    f"while loading the '{CombinatorialSupervisor.SUPERVISOR_ID}' supervisor",
                    f"The `min_match_factor` value ({v}) cannot exceed 1.0."
                )

            if v < 0.0:
                raise ErrSupervisionInvalidEngineConfig(
                    f"while loading the '{CombinatorialSupervisor.SUPERVISOR_ID}' supervisor",
                    f"The `min_match_factor` value ({v}) cannot be negative."
                )

            return v

        self._min_match_factor = self.config.get("min_match_factor", default=0.1, value_preprocessor=_min_match_factor_preprocessor)

        def _max_transformation_error_preprocessor(v: str) -> int:

            v = int(v)

            if v < 0:
                raise ErrSupervisionInvalidEngineConfig(
                    f"while loading the '{CombinatorialSupervisor.SUPERVISOR_ID}' supervisor.",
                    f"The `max_transformation_error` value ({v}) cannot be negative."
                )

            if v > 5000:
                raise ErrSupervisionInvalidEngineConfig(
                    f"while loading the '{CombinatorialSupervisor.SUPERVISOR_ID}' supervisor.",
                    f"The `max_transformation_error` value ({v}) is too high."
                )

            return v

        self._max_transformation_error = self.config.get("max_transformation_error", value_preprocessor=_max_transformation_error_preprocessor)

        def _z3_timeout_preprocessor(v: str) -> int:

            v = int(v)

            if v < 1:
                raise ErrSupervisionInvalidEngineConfig(
                    f"while loading the '{CombinatorialSupervisor.SUPERVISOR_ID}' supervisor.",
                    f"The `z3_timeout` value ({v}) cannot be negative or zero."
                )

            return v

        self._z3_timeout = self.config.get("z3_timeout", default=2500, value_preprocessor=_z3_timeout_preprocessor)

        # initialize all engine-specific values
        self._z3_context: z3.Context | None = None

        # create variables for components of the translation matrix
        self._transformation_matrix: np.ndarray | None = None

        # keys: matches (instances of Match)
        # values: z3 integer variables representing the errors for each match,
        # i.e., how consistent the match is with the affine transformation model
        self._match_weight: Dict[IMatch, z3.ArithRef] = {}

        self._minimum_weight_to_enforce: float | None = None

    def setup(self, template: ITemplate, matching_result: IMatchingResult, /) -> None:

        self._z3_context = z3.Context()

        self._transformation_matrix = np.array([
            [z3.Real("a", ctx=self._z3_context), z3.Real("b", ctx=self._z3_context)],
            [z3.Real("c", ctx=self._z3_context), z3.Real("d", ctx=self._z3_context)]
        ], dtype=z3.AstRef)

        for match in matching_result.get_all_matches():
            self._match_weight[match] = z3.Real(
                f"w_{match.keypoint_point[0]}_{match.keypoint_point[1]}_{match.target_point[0]}_{match.target_point[1]}",
                ctx=self._z3_context
            )

        # calculate the minimum weight that we need to enforce
        self._minimum_weight_to_enforce = matching_result.get_total_match_count() * self._min_match_factor

    def _get_consistency_check(self, match: IMatch, delta: np.ndarray, delta_prime: np.ndarray, /) -> z3.AstRef:
        """
        Generates a z3 formula asserting the consistency of the match with the affine linear transformation model.
        Consistency does not mean ideal matching of coordinates; rather, the template position with the affine
        transformation applied to it, must roughly be equal the target position for consistency to hold.
        """

        assert delta.shape == (2,)
        assert delta_prime.shape == (2,)
        assert match.template_point.shape == (2,)

        translated_template_point = self._transformation_matrix @ (match.template_point - delta) + delta_prime
        translated_template_point_x, translated_template_point_y = translated_template_point

        target_point_x, target_point_y = match.target_point

        return z3.And(
            translated_template_point_x - target_point_x <= self._max_transformation_error,
            target_point_x - translated_template_point_x <= self._max_transformation_error,
            translated_template_point_y - target_point_y <= self._max_transformation_error,
            target_point_y - translated_template_point_y <= self._max_transformation_error,
        )

    def supervise(self, template: ITemplate, matching_result: IMatchingResult, /) -> Iterable[SupervisionResult]:

        weights_lower_bounds = z3.And(*(self._match_weight[match] >= 0 for match in matching_result.get_all_matches()), self._z3_context)
        weights_upper_bounds = z3.And(*(self._match_weight[match] <= 1 for match in matching_result.get_all_matches()), self._z3_context)

        total_weight = z3.Sum(*(self._match_weight[match] for match in matching_result.get_all_matches()))

        solver = z3.Optimize(ctx=self._z3_context)
        solver.set("timeout", self._z3_timeout)

        solver.add(weights_lower_bounds)
        solver.add(weights_upper_bounds)
        solver.add(total_weight >= self._minimum_weight_to_enforce)

        solver.maximize(total_weight)

        for keypoint in template.keypoints:
            keypoint_matches: List[IMatch] = list(matching_result.get_matches_for_keypoint(keypoint.identifier))

            if len(keypoint_matches) == 0:
                continue

            # TODO: think whether this is a good algorithm design decision, and improve it if not
            anchor_match: IMatch = random.choice(keypoint_matches)

            delta = anchor_match.template_point
            delta_prime = anchor_match.target_point

            solver.push()

            for match in matching_result.get_all_matches():
                solver.add(z3.Implies(
                    self._match_weight[match] > 0,
                    # consistency check
                    self._get_consistency_check(match, delta, delta_prime),
                    ctx=self._z3_context
                ))

            result = solver.check()

            if result == z3.unsat:
                get_internal_afi().warn(Verbosity.INFO_VERBOSE, "Could not satisfy the imposed constraints.")
                solver.pop()
                continue

            if result == z3.unknown:
                get_internal_afi().warn(Verbosity.INFO_VERBOSE, "Could not decide the satifiability of the imposed constraints.")
                solver.pop()
                continue

            assert result == z3.sat

            model = solver.model()
            model_evaluator = np.vectorize(lambda var: model.eval(var, model_completion=True).as_fraction())  # noqa: B023

            # extract total weight and maximization target from the model
            model_total_weight = float(model_evaluator(total_weight))

            # extract transformation matrix from model
            transformation_matrix = model_evaluator(self._transformation_matrix)

            """
            delta = np.array([1217, 109], dtype=int)
            delta_prime = np.array([785, 374], dtype=int)
            transformation_matrix = np.array([[0.57689122, 0.01245905], [-0.0564908, 0.59782386]], dtype=float)
            """

            _result = SupervisionResult(
                delta=delta,
                delta_prime=delta_prime,
                transformation_matrix=transformation_matrix,
                score=model_total_weight
            )

            for match in matching_result.get_all_matches():
                match_weight = model_evaluator(self._match_weight[match])
                _result.set_match_weight(match, match_weight)

            yield _result

            solver.pop()



from __future__ import annotations

from typing import TYPE_CHECKING

import cv2
import numpy as np

from officialeye._internal.context.singleton import get_internal_context
from officialeye._internal.template.schema.loader import load_template

if TYPE_CHECKING:
    from officialeye._internal.template.external_supervision_result import ExternalSupervisionResult
    from officialeye._internal.template.internal_supervision_result import InternalSupervisionResult


def template_detect(template_path: str, /, *, target_path: str, **kwargs) -> ExternalSupervisionResult:

    from officialeye._internal.template.external_supervision_result import ExternalSupervisionResult

    with get_internal_context().setup(**kwargs):
        template = load_template(template_path)

        target: np.ndarray = cv2.imread(target_path, cv2.IMREAD_COLOR)

        internal_supervision_result: InternalSupervisionResult = template.do_detect(target)

        return ExternalSupervisionResult(internal_supervision_result)



from __future__ import annotations

from typing import TYPE_CHECKING

import cv2
import numpy as np

# noinspection PyProtectedMember
from officialeye._api.mutator import Mutator

if TYPE_CHECKING:
    from officialeye.types import ConfigDict


class GrayscaleMutator(Mutator):

    MUTATOR_ID = "grayscale"

    def __init__(self, config: ConfigDict, /):
        super().__init__(GrayscaleMutator.MUTATOR_ID, config)

    def mutate(self, img: np.ndarray, /) -> np.ndarray:
        return cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

import numpy as np

# noinspection PyProtectedMember
from officialeye._api_builtins.mutator.crop import CropMutator

if TYPE_CHECKING:
    from officialeye._api.image import IImage
    from officialeye._api.template.template import ITemplate


class IRegion(ABC):

    @property
    @abstractmethod
    def identifier(self) -> str:
        raise NotImplementedError()

    @property
    @abstractmethod
    def x(self) -> int:
        raise NotImplementedError()

    @property
    @abstractmethod
    def y(self) -> int:
        raise NotImplementedError()

    @property
    @abstractmethod
    def w(self) -> int:
        raise NotImplementedError()

    @property
    @abstractmethod
    def h(self) -> int:
        raise NotImplementedError()

    @property
    @abstractmethod
    def template(self) -> ITemplate:
        raise NotImplementedError()

    @property
    def top_left(self) -> np.ndarray:
        return np.array([self.x, self.y])

    @property
    def top_right(self) -> np.ndarray:
        return np.array([self.x + self.w, self.y])

    @property
    def bottom_left(self) -> np.ndarray:
        return np.array([self.x, self.y + self.h])

    @property
    def bottom_right(self) -> np.ndarray:
        return np.array([self.x + self.w, self.y + self.h])

    def get_image(self) -> IImage:
        _mutator = CropMutator(dict(x=self.x, y=self.y, w=self.w, h=self.h))
        _img: IImage = self.template.get_mutated_image()
        _img.apply_mutators(_mutator)
        return _img

    def insert_into_image(self, target: np.ndarray, transformed_version: np.ndarray = None):

        assert target.shape[0] == self.template.height
        assert target.shape[1] == self.template.width

        if transformed_version is None:
            transformed_version = self.get_image().load()

        target[self.y: self.y + self.h, self.x: self.x + self.w] = transformed_version

    def __str__(self) -> str:
        return f"Region '{self.identifier}'"

    def __eq__(self, o: Any) -> bool:

        if not isinstance(o, Region):
            return False

        return self.template.identifier == o.template.identifier and self.identifier == o.identifier

    def __hash__(self):
        return hash((self.template, self.identifier))


class Region(IRegion):

    def __init__(self, template: ITemplate, /, *, identifier: str, x: int, y: int, w: int, h: int):
        self._template = template
        self._identifier = identifier
        self._x = x
        self._y = y
        self._w = w
        self._h = h

    @property
    def identifier(self) -> str:
        return self._identifier

    @property
    def x(self) -> int:
        return self._x

    @property
    def y(self) -> int:
        return self._y

    @property
    def w(self) -> int:
        return self._w

    @property
    def h(self) -> int:
        return self._h

    @property
    def template(self) -> ITemplate:
        return self._template



from typing import Tuple

import cv2
import numpy as np

# noinspection PyProtectedMember
from officialeye._api.template.feature import IFeature

# noinspection PyProtectedMember
from officialeye._api.template.keypoint import IKeypoint

# noinspection PyProtectedMember
from officialeye._api.template.region import IRegion

_LABEL_COLOR_DEFAULT = (0, 0, 0xff)
_VISUALIZATION_SCALE_COEFF = 1.0 / 1400.0

_FEATURE_RECT_COLOR = (0, 0xff, 0)
_KEYPOINT_RECT_COLOR = (0, 0, 0xff)


def visualize_region(region: IRegion, img: np.ndarray, /, *, rect_color: Tuple[int, int, int], label_color=_LABEL_COLOR_DEFAULT) -> np.ndarray:

    img = cv2.rectangle(img, (region.x, region.y), (region.x + region.w, region.y + region.h), rect_color, 4)

    label_origin = (
        region.x + int(10 * img.shape[0] * _VISUALIZATION_SCALE_COEFF),
        region.y + int(30 * img.shape[0] * _VISUALIZATION_SCALE_COEFF)
    )

    font_scale = img.shape[0] * _VISUALIZATION_SCALE_COEFF
    img = cv2.putText(
        img,
        region.identifier,
        label_origin,
        cv2.FONT_HERSHEY_SIMPLEX,
        font_scale,
        label_color,
        int(2 * img.shape[0] * _VISUALIZATION_SCALE_COEFF),
        cv2.LINE_AA
    )

    return img


def visualize_feature(feature: IFeature, img: np.ndarray, /) -> np.ndarray:
    return visualize_region(feature, img, rect_color=_FEATURE_RECT_COLOR)


def visualize_keypoint(keypoint: IKeypoint, img: np.ndarray, /) -> np.ndarray:
    return visualize_region(keypoint, img, rect_color=_KEYPOINT_RECT_COLOR)



"""
Module providing an API for all OfficialEye's document detection tools.
"""

# disable unused imports ruff check
# ruff: noqa: F401

# noinspection PyProtectedMember,PyUnresolvedReferences
from officialeye._api.detection import detect



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, Iterable, List

import cv2
import numpy as np

# noinspection PyProtectedMember
from officialeye._api.template.keypoint import IKeypoint

# noinspection PyProtectedMember
from officialeye._api.template.match import IMatch, Match

# noinspection PyProtectedMember
from officialeye._api.template.matcher import Matcher
from officialeye.error.errors.matching import ErrMatchingInvalidEngineConfig

if TYPE_CHECKING:
    # noinspection PyProtectedMember
    from officialeye._api.template.template import ITemplate
    from officialeye.types import ConfigDict


def _preprocess_sensitivity(value: str, /) -> float:

    value = float(value)

    if value < 0.0:
        raise ErrMatchingInvalidEngineConfig(
            f"while loading the '{SiftFlannMatcher.MATCHER_ID}' keypoint matcher",
            f"The `sensitivity` value ({value}) cannot be negative."
        )

    if value > 1.0:
        raise ErrMatchingInvalidEngineConfig(
            f"while loading the '{SiftFlannMatcher.MATCHER_ID}' keypoint matcher",
            f"The `sensitivity` value ({value}) cannot exceed 1.0."
        )

    return value


class SiftFlannMatcher(Matcher):

    MATCHER_ID = "sift_flann"

    def __init__(self, config_dict: ConfigDict, /):
        super().__init__(SiftFlannMatcher.MATCHER_ID, config_dict)

        self._sensitivity = self.config.get("sensitivity", default=0.7, value_preprocessor=_preprocess_sensitivity)

        self._img: np.ndarray | None = None
        self._sift = None

        self._keypoints_target = None
        self._destination_target = None
        self._template: ITemplate | None = None
        self._matches: Dict[IKeypoint, List[Match]] | None = {}

    def setup(self, target: np.ndarray, template: ITemplate, /) -> None:

        self._img = cv2.cvtColor(target, cv2.COLOR_BGR2GRAY)

        # initialize the SIFT engine in CV2
        # noinspection PyUnresolvedReferences
        self._sift = cv2.SIFT_create()

        # pre-compute the sift keypoints in the target image
        self._keypoints_target, self._destination_target = self._sift.detectAndCompute(self._img, None)

        self._template = template

        self._matches = {}

    def match(self, keypoint: IKeypoint, /) -> None:

        _original_pattern_image = keypoint.get_image().load()

        pattern = cv2.cvtColor(_original_pattern_image, cv2.COLOR_BGR2GRAY)

        keypoints_pattern, destination_pattern = self._sift.detectAndCompute(pattern, None)

        index_params = {
            "algorithm": 1,
            "trees": 5
        }

        search_params = {
            "checks": 50
        }

        flann = cv2.FlannBasedMatcher(index_params, search_params)
        matches = flann.knnMatch(destination_pattern, self._destination_target, k=2)

        # we need to draw only good matches, so create a mask
        matches_mask = [[0, 0] for _ in range(len(matches))]

        result: List[Match] = []

        # filter matches
        for i, (m, n) in enumerate(matches):

            if m.distance >= self._sensitivity * n.distance:
                continue

            matches_mask[i] = [1, 0]

            pattern_point = keypoints_pattern[m.queryIdx].pt
            target_point = self._keypoints_target[m.trainIdx].pt

            # TODO: maybe one should consider rounding values here, instead of simply stripping the floating-point part
            pattern_point_vec = np.array(pattern_point, dtype=int)
            target_point_vec = np.array(target_point, dtype=int)

            match = Match(
                self._template,
                keypoint,
                keypoint_point=pattern_point_vec,
                target_point=target_point_vec
            )

            match.set_score(self._sensitivity * n.distance - m.distance)

            result.append(match)

        # TODO: visualization generation
        """
        # noinspection PyTypeChecker
        debug_image = cv2.drawMatchesKnn(
            pattern,
            keypoints_pattern,
            self._img,
            self._keypoints_target,
            matches,
            None,
            matchColor=(0, 0xff, 0),
            singlePointColor=(0xff, 0, 0),
            matchesMask=matches_mask,
            flags=cv2.DrawMatchesFlags_DEFAULT
        )

        cv2.imwrite(f"test_{keypoint.identifier}.png", debug_image)
        """

        assert keypoint not in self._matches
        self._matches[keypoint] = result

    def get_matches_for_keypoint(self, keypoint: IKeypoint, /) -> Iterable[IMatch]:
        assert keypoint in self._matches
        return self._matches[keypoint]



from __future__ import annotations

from abc import ABC
from typing import TYPE_CHECKING, Dict

# noinspection PyProtectedMember
from officialeye._api.template.region import IRegion
from officialeye._internal.context.singleton import get_internal_context

if TYPE_CHECKING:
    # noinspection PyProtectedMember
    from officialeye._api.template.template_interface import ITemplate
    from officialeye._internal.template.external_template import ExternalTemplate
    from officialeye._internal.template.internal_template import InternalTemplate


class SharedRegion(IRegion, ABC):

    def __init__(self, /, *, identifier: str, x: int, y: int, w: int, h: int):
        self._identifier = identifier
        self._x = x
        self._y = y
        self._w = w
        self._h = h

    @property
    def identifier(self) -> str:
        return self._identifier

    @property
    def x(self) -> int:
        return self._x

    @property
    def y(self) -> int:
        return self._y

    @property
    def w(self) -> int:
        return self._w

    @property
    def h(self) -> int:
        return self._h


class InternalRegion(SharedRegion, ABC):

    def __init__(self, template_id: str, region_dict: Dict[str, any], /):
        super().__init__(
            identifier=str(region_dict["id"]),
            x=int(region_dict["x"]),
            y=int(region_dict["y"]),
            w=int(region_dict["w"]),
            h=int(region_dict["h"])
        )

        self._template_id = template_id

    @property
    def template(self) -> InternalTemplate:
        return get_internal_context().get_template(self._template_id)


class ExternalRegion(SharedRegion, ABC):

    def __init__(self, internal_region: InternalRegion, external_template: ExternalTemplate, /):
        super().__init__(
            identifier=internal_region.identifier,
            x=internal_region.x,
            y=internal_region.y,
            w=internal_region.w,
            h=internal_region.h
        )

        self._external_template = external_template

    @property
    def template(self) -> ITemplate:
        return self._external_template



from __future__ import annotations

import tempfile
from typing import TYPE_CHECKING

import cv2
import numpy as np

# noinspection PyProtectedMember
from officialeye._api.template.interpretation import Interpretation

if TYPE_CHECKING:
    # noinspection PyProtectedMember
    from officialeye._api.template.feature import IFeature
    from officialeye.types import ConfigDict, FeatureInterpretation


class FileTempInterpretation(Interpretation):

    INTERPRETATION_ID = "file_temp"

    def __init__(self, config_dict: ConfigDict, /):
        super().__init__(FileTempInterpretation.INTERPRETATION_ID, config_dict)

        self._format = self.config.get("format", default="png", value_preprocessor=str)

    def interpret(self, feature_img: np.ndarray, feature: IFeature, /) -> FeatureInterpretation:

        with tempfile.NamedTemporaryFile(prefix="officialeye_", suffix=f".{self._format}", delete=False) as fp:
            fp.close()

        cv2.imwrite(fp.name, feature_img)

        return fp.name



from __future__ import annotations

from abc import ABC
from typing import TYPE_CHECKING, Dict, Iterable, List

# noinspection PyProtectedMember
from officialeye._api.template.match import IMatch

# noinspection PyProtectedMember
from officialeye._api.template.matching_result import IMatchingResult
from officialeye._internal.context.singleton import get_internal_afi
from officialeye._internal.feedback.verbosity import Verbosity
from officialeye.error.errors.matching import ErrMatchingMatchCountOutOfBounds

if TYPE_CHECKING:
    # noinspection PyProtectedMember
    from officialeye._api.template.template_interface import ITemplate


class SharedMatchingResult(IMatchingResult, ABC):
    """
    This class contains all the logic of the matching result instance, irrespective of whether we use the internal or the external representation.
    The parent process uses the internal representation, whereas the external representation is only for the child process.
    This class represents the aspects that both representations have in common.
    Therefore, it is important that this class operates only on the interface level and is picklable.
    """

    def __init__(self, template: ITemplate, /):
        # keys: keypoint ids
        # values: matches with this keypoint
        self._matches_dict: Dict[str, List[IMatch]] = {}

        for keypoint in template.keypoints:
            self._matches_dict[keypoint.identifier] = []

    def remove_all_matches(self):
        self._matches_dict = {}

    def add_match(self, match: IMatch, /):
        assert match.keypoint.identifier in self._matches_dict
        self._matches_dict[match.keypoint.identifier].append(match)

    def get_all_matches(self) -> Iterable[IMatch]:
        for keypoint_id in self._matches_dict:
            for match in self._matches_dict[keypoint_id]:
                yield match

    def get_total_match_count(self) -> int:
        match_count = 0
        for keypoint_id in self._matches_dict:
            match_count += len(self._matches_dict[keypoint_id])
        return match_count

    def get_keypoint_ids(self) -> Iterable[str]:
        for keypoint_id in self._matches_dict:
            yield keypoint_id

    def get_matches_for_keypoint(self, keypoint_id: str, /) -> Iterable[IMatch]:
        for match in self._matches_dict[keypoint_id]:
            yield match

    def validate(self):

        get_internal_afi().info(Verbosity.DEBUG, "Validating the keypoint matching result.")

        assert len(self._matches_dict) > 0

        total_match_count = 0

        # verify that for every keypoint, it has been matched a number of times that is in the desired bounds
        for keypoint_id in self._matches_dict:
            keypoint = self.template.get_keypoint(keypoint_id)

            keypoint_matches_min = keypoint.matches_min
            keypoint_matches_max = keypoint.matches_max

            keypoint_matches_count = len(self._matches_dict[keypoint_id])

            if keypoint_matches_count < keypoint_matches_min:
                raise ErrMatchingMatchCountOutOfBounds(
                    f"while checking that keypoint '{keypoint_id}' of template '{self.template.identifier}' "
                    f"has been matched a sufficient number of times",
                    f"Expected at least {keypoint_matches_min} matches, got {keypoint_matches_count}"
                )

            if keypoint_matches_count > keypoint_matches_max:

                get_internal_afi().info(
                    Verbosity.INFO_VERBOSE,
                    f"Keypoint '{keypoint_id}' of template '{self.template.identifier}' has too many matches "
                    f"(matches: {keypoint_matches_count} max: {keypoint_matches_max}). Cherry-picking the best matches.")
                # cherry-pick the best matches
                self._matches_dict[keypoint_id] = sorted(self._matches_dict[keypoint_id])[-keypoint_matches_max:]
                keypoint_matches_count = keypoint_matches_max
            else:
                get_internal_afi().info(
                    Verbosity.INFO_VERBOSE,
                    f"Keypoint '{keypoint_id}' of template '{self.template.identifier}' has been matched {keypoint_matches_count} times "
                    f"(min: {keypoint_matches_min} max: {keypoint_matches_max})."
                )

            total_match_count += keypoint_matches_count

        assert total_match_count >= 0
        if total_match_count == 0:
            raise ErrMatchingMatchCountOutOfBounds(
                f"while checking that there has been at least one match for template '{self.template.identifier}'.",
                "There have been no matches."
            )
        elif total_match_count < 3:
            raise ErrMatchingMatchCountOutOfBounds(
                f"while checking that there has been at least three matches for template '{self.template.identifier}'.",
                "There have been less than three matches."
            )



import enum
from concurrent.futures import Future

# noinspection PyProtectedMember
from multiprocessing.connection import Connection
from types import TracebackType
from typing import Any

# noinspection PyProtectedMember
from officialeye._internal.feedback.abstract import AbstractFeedbackInterface

# noinspection PyProtectedMember
from officialeye._internal.feedback.verbosity import Verbosity


class IPCMessageType(enum.IntEnum):
    ECHO = 0
    INFO = 1
    WARN = 2
    ERROR = 3

    UPDATE_STATUS = 4
    # means that the task is completed in any way, including throwing an exception
    TASK_DONE = 5


class InternalFeedbackInterface(AbstractFeedbackInterface):

    def __init__(self, verbosity: Verbosity, child_id: int, tx: Connection, /):
        super().__init__(verbosity)

        assert tx is not None

        self._child_id = child_id
        self._tx = tx

    def get_child_id(self) -> int:
        return self._child_id

    def _send_ipc_message(self, message_type: IPCMessageType, *args, **kwargs):
        ipc_message = (message_type, args, kwargs)
        self._tx.send(ipc_message)

    def echo(self, *args: Any, **kwargs: Any) -> None:
        self._send_ipc_message(IPCMessageType.ECHO, *args, **kwargs)

    def info(self, *args: Any, **kwargs: Any) -> None:
        self._send_ipc_message(IPCMessageType.INFO, *args, **kwargs)

    def warn(self, *args: Any, **kwargs: Any) -> None:
        self._send_ipc_message(IPCMessageType.WARN, *args, **kwargs)

    def error(self, *args: Any, **kwargs: Any) -> None:
        self._send_ipc_message(IPCMessageType.ERROR, *args, **kwargs)

    def update_status(self, new_status_text: str, /) -> None:
        self._send_ipc_message(IPCMessageType.UPDATE_STATUS, new_status_text)

    def dispose(self, exception_type: any = None, exception_value: BaseException | None = None, traceback: TracebackType | None = None) -> None:

        task_done_successfully: bool = exception_value is None

        self._send_ipc_message(IPCMessageType.TASK_DONE, task_done_successfully)

        self._tx.close()

    def fork(self, description: str, /) -> AbstractFeedbackInterface:
        # the internal feedback interface isn't meant to be forked
        raise NotImplementedError()

    def join(self, child: AbstractFeedbackInterface, future: Future, /) -> None:
        # the internal feedback interface isn't meant to be forked, so it cannot be joined either
        raise NotImplementedError()



from __future__ import annotations

from typing import TYPE_CHECKING, Dict

from officialeye import Context

# noinspection PyProtectedMember
from officialeye._api.template.feature import IFeature

# noinspection PyProtectedMember
from officialeye._api.template.interpretation_result import IInterpretationResult

# noinspection PyProtectedMember
from officialeye._api.template.template_interface import ITemplate
from officialeye._internal.api_implementation import IApiInterfaceImplementation
from officialeye._internal.template.external_template import ExternalTemplate

if TYPE_CHECKING:
    from officialeye._internal.template.internal_template import InternalTemplate
    from officialeye.types import FeatureInterpretation


class ExternalInterpretationResult(IInterpretationResult, IApiInterfaceImplementation):

    def __init__(self, template: InternalTemplate, feature_interpretations: Dict[str, FeatureInterpretation], /):
        self._template = ExternalTemplate(template)
        self._feature_interpretation = feature_interpretations

    @property
    def template(self) -> ITemplate:
        return self._template

    def get_feature_interpretation(self, feature: IFeature, /) -> FeatureInterpretation:

        if feature.identifier in self._feature_interpretation:
            return self._feature_interpretation[feature.identifier]

        return None

    def set_api_context(self, context: Context, /) -> None:
        self._template.set_api_context(context)

    def clear_api_context(self) -> None:
        self._template.clear_api_context()



from __future__ import annotations

from typing import TYPE_CHECKING, Iterable

import numpy as np

# noinspection PyProtectedMember
from officialeye._api.template.matching_result import IMatchingResult

# noinspection PyProtectedMember
from officialeye._api.template.supervision_result import SupervisionResult

# noinspection PyProtectedMember
from officialeye._api.template.supervisor import Supervisor

# noinspection PyProtectedMember
from officialeye._api.template.template_interface import ITemplate

if TYPE_CHECKING:
    from officialeye.types import ConfigDict

_IND_A = 0
_IND_B = 1
_IND_C = 2
_IND_D = 3


class LeastSquaresRegressionSupervisor(Supervisor):

    SUPERVISOR_ID = "least_squares_regression"

    def __init__(self, config_dict: ConfigDict, /):
        super().__init__(LeastSquaresRegressionSupervisor.SUPERVISOR_ID, config_dict)

    def setup(self, template: ITemplate, matching_result: IMatchingResult, /) -> None:
        pass

    def supervise(self, template: ITemplate, matching_result: IMatchingResult, /) -> Iterable[SupervisionResult]:

        match_count = matching_result.get_total_match_count()

        for anchor_match in matching_result.get_all_matches():

            delta = anchor_match.template_point
            delta_prime = anchor_match.target_point

            matrix = np.zeros((match_count << 1, 4), dtype=np.float64)
            rhs = np.zeros(match_count << 1, dtype=np.float64)

            for i, match in enumerate(matching_result.get_all_matches()):
                first_constraint_id = i << 1
                second_constraint_id = first_constraint_id + 1

                s = match.template_point
                d = match.target_point

                matrix[first_constraint_id][_IND_A] = s[0] - delta[0]
                matrix[first_constraint_id][_IND_B] = s[1] - delta[1]
                rhs[first_constraint_id] = d[0] - delta_prime[0]

                matrix[second_constraint_id][_IND_C] = s[0] - delta[0]
                matrix[second_constraint_id][_IND_D] = s[1] - delta[1]
                rhs[second_constraint_id] = d[1] - delta_prime[1]

            regression_matrix = matrix.T @ matrix
            regression_matrix = np.linalg.inv(regression_matrix)
            rhs_applied = matrix.T @ rhs
            x = regression_matrix @ rhs_applied

            transformation_matrix = np.array([
                [x[_IND_A], x[_IND_B]],
                [x[_IND_C], x[_IND_D]]
            ])

            _result = SupervisionResult(
                delta=delta,
                delta_prime=delta_prime,
                transformation_matrix=transformation_matrix
            )

            yield _result



from __future__ import annotations

from concurrent.futures import Future
from types import TracebackType
from typing import TYPE_CHECKING, Any

from officialeye._internal.feedback.abstract import AbstractFeedbackInterface
from officialeye._internal.feedback.verbosity import Verbosity

if TYPE_CHECKING:
    from officialeye._internal._types import RichProtocol


class DummyFeedbackInterface(AbstractFeedbackInterface):

    def __init__(self, /):
        super().__init__(Verbosity.QUIET)

    def echo(self, verbosity: Verbosity, message: str | RichProtocol = "", /, *, err: bool = False, **kwargs: Any) -> None:
        pass

    def info(self, verbosity: Verbosity, message: str, /) -> None:
        pass

    def warn(self, verbosity: Verbosity, message: str, /) -> None:
        pass

    def error(self, verbosity: Verbosity, message: str, /) -> None:
        pass

    def update_status(self, new_status_text: str, /) -> None:
        pass

    def dispose(self, exception_type: any = None, exception_value: BaseException | None = None, traceback: TracebackType | None = None) -> None:
        pass

    def fork(self, description: str, /) -> AbstractFeedbackInterface:
        return DummyFeedbackInterface()

    def join(self, child: AbstractFeedbackInterface, future: Future, /) -> None:
        pass





from __future__ import annotations

from abc import ABC, abstractmethod
from concurrent.futures import Future
from types import TracebackType
from typing import TYPE_CHECKING, Any

from officialeye._internal.feedback.verbosity import Verbosity

if TYPE_CHECKING:
    from officialeye._internal._types import RichProtocol


class AbstractFeedbackInterface(ABC):

    def __init__(self, verbosity: Verbosity, /):
        self._verbosity = verbosity

    @abstractmethod
    def echo(
        self,
        verbosity: Verbosity,
        message: str | RichProtocol = "", /, *,
        err: bool = False,
        **kwargs: Any
    ) -> None:
        raise NotImplementedError()

    @abstractmethod
    def info(self, verbosity: Verbosity, message: str, /) -> None:
        raise NotImplementedError()

    @abstractmethod
    def warn(self, verbosity: Verbosity, message: str, /) -> None:
        raise NotImplementedError()

    @abstractmethod
    def error(self, verbosity: Verbosity, message: str, /) -> None:
        raise NotImplementedError()

    @abstractmethod
    def update_status(self, new_status_text: str, /) -> None:
        raise NotImplementedError()

    @abstractmethod
    def dispose(self, exception_type: any = None, exception_value: BaseException | None = None, traceback: TracebackType | None = None) -> None:
        raise NotImplementedError()

    @abstractmethod
    def fork(self, description: str, /) -> AbstractFeedbackInterface:
        raise NotImplementedError()

    @abstractmethod
    def join(self, child: AbstractFeedbackInterface, future: Future, /) -> None:
        raise NotImplementedError()



from officialeye._internal.context.singleton import get_internal_context
from officialeye._internal.template.external_template import ExternalTemplate
from officialeye._internal.template.schema.loader import load_template


def template_load(template_path: str, /, **kwargs) -> ExternalTemplate:

    with get_internal_context().setup(**kwargs):
        template = load_template(template_path)
        return ExternalTemplate(template)



"""
OfficialEye CLI frontend main entry point.
"""

from typing import List

import click

from officialeye.__version__ import __github_full_url__, __github_url__, __version__
from officialeye._cli.context import CLIContext
from officialeye._cli.create import do_create
from officialeye._cli.run import do_run
from officialeye._cli.show import do_show
from officialeye._cli.test import do_test
from officialeye._cli.ui import Verbosity

_context = CLIContext()


@click.group()
@click.option("-d", "--debug", is_flag=True, show_default=True, default=False, help="Enable debug mode.")
@click.option("--edir", type=click.Path(exists=True, file_okay=True, readable=True), help="Specify export directory.")
@click.option("-q", "--quiet", is_flag=True, show_default=True, default=False, help="Disable standard output messages.")
@click.option("-v", "--verbose", is_flag=True, show_default=True, default=False, help="Enable verbose logging.")
@click.option("-dl", "--disable-logo", is_flag=True, show_default=True, default=False, help="Disable the officialeye logo.")
@click.option("-re", "--raw-errors", is_flag=True, show_default=False, default=False, help="Do not handle errors.")
def main(debug: bool, edir: str, quiet: bool, verbose: bool, disable_logo: bool, raw_errors: bool):
    global _context

    # configure context
    if quiet:
        verbosity = Verbosity.QUIET
    elif debug:
        if verbose:
            verbosity = Verbosity.DEBUG_VERBOSE
        else:
            verbosity = Verbosity.DEBUG
    else:
        # info verbosity
        if verbose:
            verbosity = Verbosity.INFO_VERBOSE
        else:
            verbosity = Verbosity.INFO

    _context.set_params(
        export_directory=edir,
        handle_exceptions=not raw_errors,
        verbosity=verbosity,
        disable_logo=disable_logo
    )


# noinspection PyShadowingBuiltins
@click.command()
@click.argument("template_path", type=click.Path(exists=False, file_okay=True, readable=True, writable=True))
@click.argument("template_image", type=click.Path(exists=True, file_okay=True, readable=True, writable=False))
@click.option("--id", type=str, show_default=False, default="example", help="Specify the template identifier.")
@click.option("--name", type=str, show_default=False, default="Example", help="Specify the template name.")
@click.option("--force", is_flag=True, show_default=True, default=False, help="Create missing directories and overwrite file.")
def create(template_path: str, template_image: str, id: str, name: str, force: bool):
    """Creates a new template configuration file at the specified path."""

    global _context

    with _context as context:
        context.print_logo()

        do_create(
            context,
            template_path=template_path,
            template_image=template_image,
            template_id=id,
            template_name=name,
            force_mode=force
        )


@click.command()
@click.argument("template_path", type=click.Path(exists=True, file_okay=True, readable=True))
@click.option("--hide-features", is_flag=True, show_default=False, default=False, help="Do not visualize the locations of features.")
@click.option("--hide-keypoints", is_flag=True, show_default=False, default=False, help="Do not visualize the locations of keypoints.")
def show(template_path: str, hide_features: bool, hide_keypoints: bool):
    """Exports template as an image with features visualized."""

    global _context

    with _context as context:
        do_show(context, template_path=template_path, hide_features=hide_features, hide_keypoints=hide_keypoints)


@click.command()
@click.argument("target_path", type=click.Path(exists=True, file_okay=True, readable=True))
@click.argument("template_paths", type=click.Path(exists=True, file_okay=True, readable=True), nargs=-1)
@click.option("--show-features", is_flag=True, show_default=False, default=False, help="Visualize the locations of features.")
def test(target_path: str, template_paths: List[str], show_features: bool):
    """Visualizes the analysis of an image using one or more templates."""

    global _context

    with _context as context:
        do_test(
            context,
            target_path=target_path,
            template_paths=template_paths,
            show_features=show_features
        )


@click.command()
@click.argument("target_path", type=click.Path(exists=True, file_okay=True, readable=True))
@click.argument("template_paths", type=click.Path(exists=True, file_okay=True, readable=True), nargs=-1)
@click.option("--interpret", type=click.Path(exists=True, file_okay=True, readable=True),
              default=None, help="Use the image at the specified path to run the interpretation phase.")
@click.option("--visualize", is_flag=True, show_default=False, default=False, help="Generate visualizations of intermediate steps.")
def run(target_path: str, template_paths: List[str], interpret: str | None, visualize: bool):
    """Applies one or more templates to an image."""

    global _context

    # TODO: think whether this is a good design choice
    _context.set_params(visualization_generation=visualize)

    with _context as context:
        do_run(
            context,
            target_path=target_path,
            template_paths=template_paths,
            interpret_path=interpret,
            visualize=visualize
        )


@click.command()
def homepage():
    """Go to the officialeye's official GitHub homepage."""

    global _context

    with _context as context:
        context.get_terminal_ui().info(Verbosity.INFO, f"GitHub: [link={__github_full_url__}]{__github_url__}[/link]")

    click.launch(__github_full_url__)


@click.command()
def version():
    """Print the version of OfficialEye."""

    global _context

    with _context as context:
        context.print_logo()
        context.get_terminal_ui().info(Verbosity.INFO, f"Version: {__version__}")


main.add_command(create)
main.add_command(show)
main.add_command(test)
main.add_command(run)
main.add_command(homepage)
main.add_command(version)

if __name__ == "__main__":
    main()
